import logging

from enum import Enum
from os import O_WRONLY, O_RDONLY
from threading import Thread, Event
from ctypes import Structure, c_uint32, c_uint64

from posix_ipc import MessageQueue, ExistentialError

from avatar2.message import RemoteMemoryReadMessage, RemoteMemoryWriteMessage


class Operation(Enum):
    READ = 0
    WRITE = 1


class RemoteMemoryReq(Structure):
    _fields_ = [
        ('id', c_uint64),
        ('pc', c_uint64),
        ('address', c_uint64),
        ('value', c_uint64),
        ('size', c_uint32),
        ('operation', c_uint32)
    ]


class RemoteMemoryResp(Structure):
    _fields_ = [
        ('id', c_uint64),
        ('value', c_uint64),
        ('success', c_uint32)
    ]


class RemoteMemoryRequestListener(Thread):
    def __init__(self, rx_queue, avatar_queue, origin):
        super(RemoteMemoryRequestListener, self).__init__()
        self._rx_queue = rx_queue
        self._avatar_queue = avatar_queue
        self._origin = origin
        self._close = Event()
        self._closed = Event()
        self._close.clear()
        self._closed.clear()
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def run(self):
        while True:
            if self._close.is_set():
                break

            request = None
            try:
                request = self._rx_queue.receive(0.5)
            except:
                continue

            req_struct = RemoteMemoryReq.from_buffer_copy(request[0])

            if Operation(req_struct.operation) == Operation.READ:
                self.log.debug(("Received RemoteMemoryRequest."
                                "Read from 0x%x at 0x%x") %
                                (req_struct.address, req_struct.pc))
                MemoryForwardMsg = RemoteMemoryReadMessage(self._origin,
                                                           req_struct.id,
                                                           req_struct.pc,
                                                           req_struct.address,
                                                           req_struct.size)
            elif Operation(req_struct.operation) == Operation.WRITE:
                self.log.debug(("Received RemoteMemoryRequest."
                                "Write to 0x%x at 0x%x") %
                                (req_struct.address, req_struct.pc))
                MemoryForwardMsg = RemoteMemoryWriteMessage(self._origin,
                                                            req_struct.id,
                                                            req_struct.pc,
                                                            req_struct.address,
                                                            req_struct.value,
                                                            req_struct.size)
            else:
                raise ValueError("Received Message with unkown operation %d" %
                                 req_struct.operation)

            self._avatar_queue.put(MemoryForwardMsg)

        self._closed.set()

    def stop(self):
        self._close.set()
        self._closed.wait()


class RemoteMemoryProtocol(object):
    """
    This class listens to memoryforward requests and lifts them to avatar
    messages. Likewise it can be directed to emit memoryforward-response
    messages

    :param rx_queue_name: Name of the queue for receiving
    :param tx_queue_name: Name of the queue for sending
    :param avatar_queue:  Queue to dispatch received requests to
    :param origin:        Reference to the Target utilizing this protocol

    """

    def __init__(self, rx_queue_name, tx_queue_name,
                 avatar_queue, origin=None):
        self._rx_queue = None
        self._tx_queue = None
        self._rx_listener = None

        self.rx_queue_name = rx_queue_name
        self.tx_queue_name = tx_queue_name
        self._avatar_queue = avatar_queue
        self._origin = origin

        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def connect(self):
        """
        Connect to the message queues for remote memory

        :return True on success, else False
        """
        try:
            self._rx_queue = MessageQueue(self.rx_queue_name, flags=O_RDONLY,
                                          read=True, write=False)
        except Exception as e:
            self.log.exception("Unable to create rx_queue:")
            return False

        try:
            self._tx_queue = MessageQueue(self.tx_queue_name, flags=O_WRONLY,
                                          read=False, write=True)
        except Exception as e:
            self.log.exception("Unable to create tx_queue:")
            self._rx_queue.close()
            return False
        self._rx_listener = RemoteMemoryRequestListener(self._rx_queue,
                                                        self._avatar_queue,
                                                        self._origin)
        self._rx_listener.daemon = True
        self._rx_listener.start()
        self.log.info("Successfully connected rmp")
        return True

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        if self._rx_listener:
            self._rx_listener.stop()
            self._rx_listener = None
        if self._rx_queue:
            try:
                self._rx_queue.unlink()
                self._rx_queue.close()
                self._rx_queue = None
            except ExistentialError:
                self.log.warning("Tried to close/unlink non existent rx_queue")
        if self._tx_queue:
            try:
                self._tx_queue.unlink()
                self._tx_queue.close()
                self._tx_queue = None
            except ExistentialError:
                self.log.warning("Tried to close/unlink non existent tx_queue")

    def send_response(self, id, value, success):
        response = RemoteMemoryResp(id, value, success)
        try:
            self._tx_queue.send(response)
            self.log.debug("Send RemoteMemoryResponse with id %d, %x" % (id, value))
            return True
        except Exception as e:
            self.log.error("Unable to send response: %s" % e)
            return False



from avatar2.targets import Target, TargetStates
from avatar2.protocols.unicorn_protocol import UnicornProtocol


class UnicornTarget(Target):
    def __init__(self, avatar, **kwargs):
        super(UnicornTarget, self).__init__(avatar, **kwargs)

    def init(self):
        proto = UnicornProtocol(self.avatar, arch=self._arch, origin=self)
        self.protocols.set_all(proto)
        self.protocols.remote_memory = proto

    def set_breakpoint(self, line, **kwargs):
        if 'hardware' not in kwargs:
            kwargs['hardware'] = True
        self.protocols.execution.set_breakpoint(line, **kwargs)


    def add_hook(self, event, hook):
        # Allows the user to set a unicorn hook
        self.protocols.execution.uc.hook_add(event, hook)



# Author: Davide Balzarotti <davide.balzarotti@eurecom.fr>
# Creation Date: 04-04-2017

import random
import threading
import time

from avatar2.message import RemoteMemoryReadMessage, BreakpointHitMessage, UpdateStateMessage
from avatar2.targets import Target, TargetStates


class _TargetThread(threading.Thread):
    """Thread that mimics a running target"""

    def __init__(self, target):
        threading.Thread.__init__(self)
        self.target = target
        self.please_stop = False
        self.steps = 0

    def run(self):
        self.please_stop = False
        # Loops until someone (the Dummy Target) tells me to stop by 
        # externally setting the "please_stop" variable to True 
        while not self.please_stop:
            self.target.log.info("Dummy target doing Nothing..")
            time.sleep(1)
            self.steps += 1
            # 10% chances of triggering a breakpoint
            if random.randint(0, 100) < 10:
                # If there are not breakpoints set, continue
                if len(self.target.bp) == 0:
                    continue
                # Randomly pick one of the breakpoints
                addr = random.choice(self.target.bp)
                self.target.log.info("Taking a break..")
                # Add a message to the Avatar queue to trigger the
                # breakpoint 
                self.target.avatar.queue.put(BreakpointHitMessage(self.target, 1, addr))
                break
            # 90% chances of reading from a forwarded memory address
            else:
                # Randomly pick one forwarded range
                mem_range = random.choice(self.target.mranges)
                # Randomly pick an address in the range
                addr = random.randint(mem_range[0], mem_range[1] - 4)
                # Add a message in the Avatar queue to read the value at
                # that address
                self.target.avatar.queue.put(RemoteMemoryReadMessage(self.target, 55, addr, 4))
        self.target.log.info("Avatar told me stop..")


class DummyTarget(Target):
    """ 
    This is a Dummy target that can be used for testing purposes.
    It simulates a device that randomly reads from forwarded memory ranges
    and triggers breakpoints.    
    """

    def __init__(self, avatar, **kwargs):
        super(DummyTarget, self).__init__(avatar, **kwargs)
        # List of breakpoints
        self.bp = []
        # List of forwarded memory ranges 
        self.mranges = []
        self.thread = None

        # Avatar will try to answer to our messages (e.g., with the value
        # the Dummy Target tries to read from memory). To handle that
        # we need a memory protocol. However, here we set the protocol to
        # ourself (its a dirty trick) and later implement the sendResponse 
        # method
        self.protocols.remote_memory = self

    # This is called by Avatar to initialize the target
    def init(self):
        self.log.info("Dummy Target Initialized and ready to rock")
        # Ack. It should actually go to INITIALIZED but then the protocol
        # should change that to STOPPED 
        self.avatar.queue.put(UpdateStateMessage(self, TargetStates.STOPPED))
        # We fetch from Avatar the list of memory ranges that are
        # configured to be forwarded
        for mem_range in self.avatar.memory_ranges:
            mem_range = mem_range.data
            if mem_range.forwarded:
                self.mranges.append((mem_range.address, mem_range.address + mem_range.size))
        self.wait()

    # If someone ones to read memory from this target, we always return
    # the same value, no matter what address it is requested
    def read_memory(*args, **kwargs):
        return 0xdeadbeef

    # This allow Avatar to answer to our memory read requests.
    # However, we do not care about it
    def send_response(self, id, value, success):
        if success:
            self.log.debug("RemoteMemoryRequest with id %d returned 0x%x" %
                           (id, value))
        else:
            self.log.warning("RemoteMemoryRequest with id %d failed" % id)

    # We let Avatar writes to our memory.. well.. at least we let it
    # believe so
    def write_memory(self, addr, size, val, *args, **kwargs):
        return True

    # We keep tracks of breakpoints
    def set_breakpoint(self, line, hardware=False, temporary=False, regex=False, condition=None, ignore_count=0,
                       thread=0):
        self.bp.append(line)

    def remove_breakpoint(self, breakpoint):
        # FIXME.. how do you remove a breakpoint?
        # sle.bp.remove(breakpoint) does not work
        pass

        # def wait(self):
        # self.thread.join()

    def cont(self):
        if self.state != TargetStates.RUNNING:
            self.avatar.queue.put(UpdateStateMessage(self, TargetStates.RUNNING))
            self.thread = _TargetThread(self)
            self.thread.daemon = True
            self.thread.start()

    def get_status(self):
        if self.thread:
            self.status.update({"state": self.state, "steps": self.thread.steps})
        else:
            self.status.update({"state": self.state, "steps": '-'})
        return self.status

    # Since we set the memory protocol to ourself, this is important to avoid 
    # an infinite recursion (otherwise by default a target would call
    # shutdown to all its protocols)
    def shutdown(self):
        pass

    def stop(self):
        if self.state == TargetStates.RUNNING:
            self.thread.please_stop = True
            self.avatar.queue.put(UpdateStateMessage(self, TargetStates.STOPPED))
        return True



from .target import Target, TargetStates, action_valid_decorator_factory
from .dummy_target import DummyTarget
from .openocd_target import *
from .gdb_target import *
from .jlink_target import *
from .qemu_target import *
from .panda_target import *
from .pypanda_target import *
from .unicorn_target import *



from types import MethodType
from threading import Thread, Event, Condition

from avatar2 import TargetStates
from avatar2.archs import ARMV7M
from avatar2.protocols.coresight import CoreSightProtocol
from avatar2.protocols.armv7m_interrupt import ARMV7MInterruptProtocol
from avatar2.targets import OpenOCDTarget, QemuTarget
from avatar2.watchmen import AFTER

from avatar2.message import RemoteInterruptEnterMessage
from avatar2.message import RemoteInterruptExitMessage
from avatar2.message import RemoteMemoryWriteMessage
from avatar2.message import BreakpointHitMessage

from avatar2.watchmen import watch


def add_protocols(self, **kwargs):
    target = kwargs['watched_target']
    if isinstance(target, OpenOCDTarget):
        target.protocols.interrupts = CoreSightProtocol(target.avatar,
                                                        target)
        # We want to remove the decorators around the read_memory function of
        # this target, to allow reading while it is running (thanks oocd)
        target.read_memory = MethodType( lambda t, *args, **kwargs:
                                        t.protocols.memory.read_memory(
                                            *args, **kwargs), target
                                       )
        # Also, let's use openocd as protocol for register and memory
        target.protocols.memory = target.protocols.monitor
        target.protocols.register = target.protocols.monitor

        
    if isinstance(target, QemuTarget):
        target.protocols.interrupts = ARMV7MInterruptProtocol(
            target, self.v7m_irq_rx_queue_name, self.v7m_irq_tx_queue_name
        )


def forward_interrupt(self, message):  # , **kwargs):
    global stawp
    target = message.origin
    target.update_state(message.state)
    self.queue.put(message)

    if isinstance(target, OpenOCDTarget):
        if message.address == message.origin.protocols.interrupts._monitor_stub_isr - 1:
            xpsr = target.read_register('xPSR')
            irq_num = xpsr & 0xff
            self.log.info("Injecting IRQ 0x%x" % irq_num)
            self._irq_dst.protocols.interrupts.inject_interrupt(irq_num)


def gontinue_execution(self, message, **kwargs):
    target = message.origin
    if message.address == message.origin.protocols.interrupts._monitor_stub_isr - 1:
        target.cont()


def enable_interrupt_forwarding(self, from_target, to_target=None,
                                disabled_irqs=None, semi_forwarding=False):
    """
    Semi forwarding is a special mode developed for pretender.
    It allows that irqs are taken from from_target and external calls to
    inject_interrupt. However, no information about to_targets irq-state is
    given back to from_target. Nevertheless, memory requests from to_target to
    from_target are forwarded.
    Confused yet? So are we, this is a huge hack.
    """
    self._irq_src = from_target
    self._irq_dst = to_target
    self._irq_semi_forwarding = semi_forwarding
    self._irq_ignore = [] if disabled_irqs is None else disabled_irqs

    self._handle_remote_interrupt_enter_message = MethodType(
        _handle_remote_interrupt_enter_message, self)
    self._handle_remote_interrupt_exit_message = MethodType(
        _handle_remote_interrupt_exit_message, self)
    self._handle_remote_memory_write_message_nvic = MethodType(
        _handle_remote_memory_write_message_nvic, self)

    self.message_handlers.update(
        {
            RemoteInterruptEnterMessage: self._handle_remote_interrupt_enter_message,
            RemoteInterruptExitMessage: self._handle_remote_interrupt_exit_message}
    )
    self.message_handlers.update(
        {
            RemoteMemoryWriteMessage: self._handle_remote_memory_write_message_nvic}
    )

    if from_target:
        from_target.protocols.interrupts.enable_interrupts()
        isr_addr = from_target.protocols.interrupts._monitor_stub_isr - 1
        self.log.info("ISR breakpoint at %#08x" % isr_addr)
        from_target.set_breakpoint(isr_addr, hardware=True)
    if to_target:
        to_target.protocols.interrupts.enable_interrupts()

    # OpenOCDProtocol does not emit breakpointhitmessages currently,
    # So we listen on state-updates and figure out the rest on our own
    # self.watchmen.add_watchman('BreakpointHit', when=AFTER,
    #                             callback=continue_execution)
    self._handle_breakpoint_handler = MethodType(forward_interrupt, self)
    self.fast_queue_listener.message_handlers.update({
        BreakpointHitMessage: self._handle_breakpoint_handler
    }
    )

    # def _fast_handle_update_state_message(self, message):
    # print message
    # message.origin.update_state(message.state)
    # self.avatar.queue.put(message)


@watch('RemoteInterruptEnter')
def _handle_remote_interrupt_enter_message(self, message):
    self._irq_dst.protocols.interrupts.send_interrupt_enter_response(message.id,
                                                                     True)
    if self._irq_src is None or self._irq_semi_forwarding is True:
        return

    
    status = self._irq_src.get_status()
    if status['state'] == TargetStates.STOPPED:
        self.log.info("Target stopped, restarting " + repr(message.origin))
        try:
            self._irq_src.cont(blocking=False)
        except:
            self.log.exception(" ")



@watch('RemoteInterruptExit')
def _handle_remote_interrupt_exit_message(self, message):
    """
    Handle an interrupt exiting properly
    If the interrupt was trigged by the hardware, we need to tell the
    interrupt that we satisified it
    :param self:
    :param message:
    :return:
    """
    if self._irq_src is not None and self._irq_semi_forwarding is False:
        # We are forwarding, make sure to forward the return
        self._irq_src.protocols.interrupts.inject_exc_return(
            message.transition_type)

    # Always ack the exit message
    self._irq_dst.protocols.interrupts.send_interrupt_exit_response(message.id,
                                                                    True)


def _handle_remote_memory_write_message_nvic(self, message):
    # NVIC address according to coresight manual
    if message.address < 0xe000e000 or message.address > 0xe000f000 or self._irq_src is None:
        return self._handle_remote_memory_write_message(message)

    # Discard writes to the vector table offset registers
    # TODO add other blacklists
    if message.address == 0xE000ED08:
        success = True
    else:
        success = self._irq_src.write_memory(message.address,
                                             message.size,
                                             message.value)

    message.origin.protocols.remote_memory.send_response(message.id, 0,
                                                         success)
    return message.id, message.value, success


def load_plugin(avatar):
    if avatar.arch != ARMV7M:
        avatar.log.error("Tried to load armv7-m interrupt plugin " +
                         "with mismatching architecture")

    avatar.v7m_irq_rx_queue_name = '/avatar_v7m_irq_rx_queue'
    avatar.v7m_irq_tx_queue_name = '/avatar_v7m_irq_tx_queue'
    avatar.enable_interrupts = MethodType(enable_interrupt_forwarding, avatar)

    avatar.watchmen.add_watchman('TargetInit', when=AFTER,
                                 callback=add_protocols)



from types import MethodType
from threading import Event
from enum import Enum

from avatar2.watchmen import AFTER, BEFORE, watch
from avatar2 import TargetStates

watched_events = {
    'OrchestrationTransitionAdd',
    'OrchestrationTransition',
    'OrchestrationStart',
    'OrchestrationResumed',
    'OrchestrationStop',
    'OrchestrationTransitionsDisabled',
    'OrchestrationTransitionsEnabled'
}


class OrchestrationStopReason(Enum):
    STOPPING_TRANSITION_HIT = 0
    UNKNOWN_BREAKPOINT_HIT = 1
    TARGET_EXITED = 2
    USER_REQUESTED = 3


class Transition(object):
    def __init__(self, address, from_target, to_target,
                 sync_regs, synced_ranges, enabled=True,
                 max_hits=0, stop=False, hw_bkpt=False):
        self.address = address
        self.from_target = from_target
        self.to_target = to_target
        self.sync_regs = sync_regs
        self.synced_ranges = synced_ranges
        self.enabled = enabled
        self.max_hits = max_hits
        self.num_hits = 0
        self.stop = stop
        self.hw_bkpt = hw_bkpt


def update_state_callback(avatar, message, **kwargs):
    if message.state == TargetStates.EXITED and \
                    message.origin == avatar.last_target:
        avatar.stop_orchestration(
            OrchestrationStopReason.TARGET_EXITED)


def transition_callback(avatar, message, **kwargs):
    from_target = message.origin
    address = message.address

    if avatar.transitions.get((address, from_target), None) is not None:
        trans = avatar.transitions[(address, from_target)]
        if trans.enabled:
            avatar.watchmen.trigger('OrchestrationTransition', BEFORE, trans)
            avatar.transfer_state(from_target, trans.to_target,
                                  trans.sync_regs, trans.synced_ranges)
            trans.num_hits += 1
            avatar.last_target = trans.to_target
            if trans.stop == True:
                avatar.stop_orchestration(
                    OrchestrationStopReason.STOPPING_TRANSITION_HIT)
            else:
                trans.to_target.cont()
            avatar.watchmen.trigger('OrchestrationTransition', AFTER, trans)
    elif avatar.orchestration_stopped.is_set() == False:
        avatar.stop_orchestration(
            OrchestrationStopReason.UNKNOWN_BREAKPOINT_HIT)


@watch('OrchestrationTransitionAdd')
def add_transition(self, address, from_target, to_target,
                   sync_regs=True, synced_ranges=None, stop=False,
                   hw_breakpoint=False):
    if synced_ranges is None:
        synced_ranges = []
    trans = Transition(address, from_target, to_target,
                       sync_regs=sync_regs,
                       synced_ranges=synced_ranges,
                       stop=stop, hw_bkpt=hw_breakpoint)

    self.transitions[(address, from_target)] = trans


@watch('OrchestrationTransitionsEnabled')
def enable_transitions(self):
    for t in self.targets.values():
        if t.state != TargetStates.STOPPED:
            raise Exception("%s has to be stopped to enable transitions" % t)
    for t in self.transitions.values():
        t.bkptno = t.from_target.set_breakpoint(t.address, hardware=t.hw_bkpt)


@watch('OrchestrationTransitionsDisabled')
def disable_transitions(self):
    for t in self.transitions.values():
        t.from_target.remove_breakpoint(t.bkptno)


def _orchestrate(self, target, blocking=True):
    self.enable_transitions()
    self.orchestration_stopped_reason = None
    self.orchestration_stopped.clear()
    self.last_target = target

    target.cont()
    if blocking == True:
        saved_handler = self.sigint_handler
        self.sigint_handler = self.stop_orchestration
        self.orchestration_stopped.wait()
        self.sigint_handler = saved_handler


@watch('OrchestrationStart')
def start_orchestration(self, force_init=False, blocking=True):
    if self.start_target is None:
        raise Exception("No starting target specified!")
    for t in self.targets.values():
        if t.state == TargetStates.CREATED or force_init:
            t.init()

    self._orchestrate(self.start_target, blocking)


@watch('OrchestrationResumed')
def resume_orchestration(self, blocking=True):
    if self.last_target == None:
        raise Exception("No Orchestration was running before!")
    self._orchestrate(self.last_target, blocking)


@watch('OrchestrationStop')
def stop_orchestration(self, reason=OrchestrationStopReason.USER_REQUESTED):
    for t in self.targets.values():
        if t.state == TargetStates.RUNNING:
            t.stop()
    self.disable_transitions()
    self.orchestration_stopped.set()
    self.orchestration_stopped_reason = reason


def load_plugin(avatar):
    avatar.transitions = {}
    avatar.orchestration_stopped = Event()
    avatar.orchestration_stopped.set()
    avatar.orchestration_stopped_reason = None

    avatar.start_target = None
    avatar.last_target = None

    avatar.watchmen.add_watch_types(watched_events)
    avatar.watchmen.add_watchman('BreakpointHit', when=AFTER,
                                 callback=transition_callback)
    avatar.watchmen.add_watchman('UpdateState', when=AFTER,
                                 callback=update_state_callback)

    avatar.add_transition = MethodType(add_transition, avatar)
    avatar.enable_transitions = MethodType(enable_transitions, avatar)
    avatar.disable_transitions = MethodType(disable_transitions, avatar)
    avatar.start_orchestration = MethodType(start_orchestration, avatar)
    avatar.resume_orchestration = MethodType(resume_orchestration, avatar)
    avatar.stop_orchestration = MethodType(stop_orchestration, avatar)
    avatar._orchestrate = MethodType(_orchestrate, avatar)



import json
from subprocess import Popen
from os.path import isfile, exists

from avatar2.protocols.gdb import GDBProtocol
from avatar2.protocols.qmp import QMPProtocol
from avatar2.protocols.remote_memory import RemoteMemoryProtocol
from avatar2.targets import Target

from avatar2.installer.config import QEMU, GDB_MULTI
from avatar2.watchmen import watch


class QemuTarget(Target):
    """"""

    def __init__(
        self,
        avatar,
        executable=None,
        cpu_model=None,
        firmware=None,
        gdb_executable=None,
        gdb_port=3333,
        additional_args=None,
        gdb_additional_args=None,
        gdb_verbose=False,
        qmp_port=3334,
        entry_address=0x00,
        log_items=None,
        log_file=None,
        system_clock_scale=None,
        **kwargs
    ):
        super(QemuTarget, self).__init__(avatar, **kwargs)

        # Qemu parameters
        self.system_clock_scale = system_clock_scale
        if hasattr(self, "executable") is False and self.__class__ == QemuTarget:
            self.executable = (
                executable
                if executable is not None
                else self._arch.get_qemu_executable()
            )
        self.fw = firmware
        self.cpu_model = cpu_model
        self.entry_address = entry_address
        self.additional_args = additional_args if additional_args else []

        # gdb parameters
        self.gdb_executable = (
            gdb_executable
            if gdb_executable is not None
            else self._arch.get_gdb_executable()
        )

        self.gdb_port = gdb_port
        self.gdb_additional_args = gdb_additional_args if gdb_additional_args else []
        self.gdb_verbose = gdb_verbose

        self.qmp_port = qmp_port

        self._process = None
        self._entry_address = entry_address
        self._memory_mapping = avatar.memory_ranges

        self._rmem_rx_queue_name = "/{:s}_rx_queue".format(self.name)
        self._rmem_tx_queue_name = "/{:s}_tx_queue".format(self.name)

        self.log_items = log_items
        self.log_file = log_file

        self.qemu_config_file = "%s/%s_conf.json" % (
            self.avatar.output_directory,
            self.name,
        )

    def assemble_cmd_line(self):
        if isfile(self.executable + self._arch.qemu_name):
            executable_name = [self.executable + self._arch.qemu_name]
        elif isfile(self.executable):
            executable_name = [self.executable]
        else:
            raise Exception(
                "Executable for %s not found: %s" % (self.name, self.executable)
            )

        machine = ["-machine", "configurable"]
        kernel = ["-kernel", self.qemu_config_file]
        gdb_option = ["-gdb", "tcp::" + str(self.gdb_port)]
        stop_on_startup = ["-S"]
        nographic = ["-nographic"]  # , "-monitor", "/dev/null"]
        qmp = ["-qmp", "tcp:127.0.0.1:%d,server,nowait" % self.qmp_port]

        cmd_line = (
            executable_name
            + machine
            + kernel
            + gdb_option
            + stop_on_startup
            + self.additional_args
            + nographic
            + qmp
        )

        if self.log_items is not None:
            if isinstance(self.log_items, str):
                log_items = ["-d", self.log_items]
            elif isinstance(self.log_items, list):
                log_items = ["-d", ",".join([i for i in self.log_items])]
            else:
                self.log.warn(
                    "Got unsupported type for log_items: %s" % type(self.log_items)
                )
                return cmd_line

            if self.log_file is not None:
                log_file = ["-D", "%s" % self.log_file]
            else:
                log_file = [
                    "-D",
                    "%s/%s_log.txt" % (self.avatar.output_directory, self.name),
                ]

            cmd_line += log_items + log_file

        return cmd_line

    def shutdown(self):
        if self._process is not None:
            self._process.terminate()
            self._process.wait()
            self._process = None
        super(QemuTarget, self).shutdown()

    def _filter_mr_kwargs(self, mr):
        """
        Return a dict with known/expected kwargs/keys from an dictified mr.
        """
        expected_keys = [
            "name",
            "address",
            "size",
            "forwarded",
            "forwarded_to",
            "is_special",
            "is_symbolic",
            "permissions",
            "bus",
            "properties",
            "qemu_name",
            "python_peripheral",
            "inline_module",
        ]
        filtered_mr = {}
        for k, v in mr.items():
            if k in expected_keys:
                continue
            filtered_mr[k] = v
        return filtered_mr

    def generate_qemu_config(self):
        """
        Generates the configuration passed to avatar-qemus configurable machine
        """
        conf_dict = self.avatar.generate_config()
        conf_dict["entry_address"] = self.entry_address
        if self.fw is not None:
            conf_dict["kernel"] = self.fw

        if self.system_clock_scale is not None:
            conf_dict["system_clock_scale"] = self.system_clock_scale

        for mr in conf_dict["memory_mapping"]:
            if mr.get("qemu_name"):
                mr["properties"] = []
                mr["bus"] = "sysbus"
                if mr["qemu_name"] in ["avatar-rmemory", "avatar-pyperipheral"]:
                    size_properties = {
                        "type": "uint32",
                        "value": mr["size"],
                        "name": "size",
                    }
                    mr["properties"].append(size_properties)
                    address_properties = {
                        "type": "uint64",
                        "value": mr["address"],
                        "name": "address",
                    }
                    mr["properties"].append(address_properties)

                if mr["qemu_name"] == "avatar-rmemory":
                    rx_queue_properties = {
                        "type": "string",
                        "value": self._rmem_rx_queue_name,
                        "name": "rx_queue_name",
                    }
                    mr["properties"].append(rx_queue_properties)
                    tx_queue_properties = {
                        "type": "string",
                        "value": self._rmem_tx_queue_name,
                        "name": "tx_queue_name",
                    }
                    mr["properties"].append(tx_queue_properties)

                if mr["qemu_name"] == "avatar-pyperipheral":
                    file_properties = {
                        "type": "string",
                        "name": "python_file",
                        "value": mr["inline_module"],
                    }
                    mr["properties"].append(file_properties)
                    class_properties = {
                        "type": "string",
                        "name": "python_class",
                        "value": mr["python_peripheral"],
                    }
                    mr["properties"].append(class_properties)
                    filtered_kwargs = self._filter_mr_kwargs(mr)
                    kwargs_properties = {
                        "type": "string",
                        "name": "python_kwargs",
                        "value": str(filtered_kwargs),
                    }
                    mr["properties"].append(kwargs_properties)

                elif mr.get("qemu_properties"):
                    if type(mr["qemu_properties"]) == list:
                        mr["properties"] += mr["qemu_properties"]
                    else:
                        mr["properties"].append(mr["qemu_properties"])
        del conf_dict["targets"]
        return conf_dict

    @watch("TargetInit")
    def init(self, cmd_line=None):
        """
        Spawns a Qemu process and connects to it
        """

        if self.cpu_model is None:
            if hasattr(self._arch, "cpu_model"):
                self.cpu_model = self.avatar.arch.cpu_model
            else:
                self.log.warning("No cpu_model specified - are you sure?")

        if cmd_line is None:
            cmd_line = self.assemble_cmd_line()

        self.avatar.save_config(
            file_name=self.qemu_config_file, config=self.generate_qemu_config()
        )

        with open(
            "%s/%s_out.txt" % (self.avatar.output_directory, self.name), "wb"
        ) as out, open(
            "%s/%s_err.txt" % (self.avatar.output_directory, self.name), "wb"
        ) as err:
            self._process = Popen(cmd_line, stdout=out, stderr=err)
        self.log.debug("QEMU command line: %s" % " ".join(cmd_line))
        self.log.info("QEMU process running")
        self._connect_protocols()

    def _connect_protocols(self):
        """
        Internal routine to connect the various protocols to a running qemu
        """

        gdb = GDBProtocol(
            gdb_executable=self.gdb_executable,
            arch=self.avatar.arch,
            verbose=self.gdb_verbose,
            additional_args=self.gdb_additional_args,
            avatar=self.avatar,
            origin=self,
        )
        qmp = QMPProtocol(self.qmp_port, origin=self)

        if "avatar-rmemory" in [
            i[2].qemu_name
            for i in self._memory_mapping.iter()
            if hasattr(i[2], "qemu_name")
        ]:
            rmp = RemoteMemoryProtocol(
                self._rmem_tx_queue_name,
                self._rmem_rx_queue_name,
                self.avatar.queue,
                self,
            )
        else:
            rmp = None

        self.protocols.set_all(gdb)
        self.protocols.monitor = qmp
        self.protocols.remote_memory = rmp

        if gdb.remote_connect(port=self.gdb_port) and qmp.connect():
            self.log.info("Connected to remote target")
        else:
            self.log.warning("Connection to remote target failed")
        if rmp:
            rmp.connect()
        self.wait()



import sys
if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

import struct
import unicorn
import logging

from threading import Thread
from collections import namedtuple

from avatar2.message import UpdateStateMessage, RemoteMemoryReadMessage, \
    RemoteMemoryWriteMessage, BreakpointHitMessage
from avatar2.targets import TargetStates
from avatar2.archs.arm import ARM


class UnicornBreakpoint(object):
    __slots__ = ('hooks', 'temporary', 'ignore_count')

    def __init__(self, hooks, temporary=False, ignore_count=0):
        self.hooks = hooks
        self.temporary = temporary
        self.ignore_count = ignore_count


UnicornWorkerEmuStartMessage = namedtuple('UnicornWorkerEmuStartMessage', ('single_step',))
UnicornWorkerUpdateStateMessage = namedtuple('UnicornWorkerUpdateStateMessage', ('state',))
UnicornWorkerBreakpointMessage = namedtuple('UnicornWorkerBreakpointMessage', ('bkptno', 'address'))


class UnicornProtocol(object):
    """Main class for the Unicorn protocol.

    :ivar uc:         the Unicorn instance
    :ivar log:        this protocol's logger
    :ivar arch:       this protocol's architecture
    :ivar pending_bp: set of pending breakpoint numbers
    """

    def __init__(self, avatar, arch=ARM, origin=None):
        """Construct the protocol, along with its Unicorn instance and worker.

        :param avatar: the Avatar object
        :param arch:   the architecture
        :param origin: the target utilizing this protocol
        """
        self.uc = unicorn.Uc(arch.unicorn_arch, arch.unicorn_mode)
        self.log = logging.getLogger((origin.log.name + '.' if origin is not None else '') +
                                     self.__class__.__name__)
        self.arch = arch
        self.pending_bp = set()
        self._avatar_queue = avatar.queue
        self._avatar_fast_queue = avatar.fast_queue
        self._origin = origin
        self._breakpoints = []
        self._rmp_queue = queue.Queue()
        self._alive = True

        for start, end, mr in avatar.memory_ranges:
            perms = unicorn.UC_PROT_NONE
            if 'r' in mr.permissions:
                perms |= unicorn.UC_PROT_READ
            if 'w' in mr.permissions:
                perms |= unicorn.UC_PROT_WRITE
            if 'x' in mr.permissions:
                perms |= unicorn.UC_PROT_EXEC

            self.uc.mem_map(start, end - start, perms=perms)

            if hasattr(mr, 'file') and mr.file is not None:
                with open(mr.file, 'rb') as data:
                    self.uc.mem_write(start, data.read())

            if mr.forwarded:
                self.uc.hook_add(unicorn.UC_HOOK_MEM_VALID, self._forward_hook,
                                 begin=start, end=end)

        self._avatar_fast_queue.put(UpdateStateMessage(self._origin, TargetStates.INITIALIZED))

        self._worker_queue = queue.Queue()
        self._worker_queue.put(UnicornWorkerUpdateStateMessage(TargetStates.STOPPED))

        self._worker = UnicornWorker(self._origin, self, self.uc, self._worker_queue,
                                     self._avatar_fast_queue)
        self._worker.start()

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        """Shutdown the protocol."""
        if self._alive:
            self._worker_queue.put(None)
            self.stop()
            self._worker.join()
            self._alive = False

    # Execution protocol

    def cont(self):
        """Continue execution."""
        self._worker_emu_start()

    def stop(self):
        """Stop execution."""
        self._worker_emu_stop()

    def step(self):
        """Execute one instruction and stop."""
        self._worker_emu_start(single_step=True)

    def set_breakpoint(self, line, hardware=True, temporary=False, regex=False, condition=None,
                       ignore_count=0, thread=0):
        """Insert a breakpoint.

        :param line:         address to break at
        :param hardware:     whether this breakpoint is hardware (ignored, always True)
        :param temporary:    whether this breakpoint is temporary (one shot)
        :param regex:        not supported
        :param condition:    not supported
        :param ignore_count: amount of times the breakpoint should be ignored before firing
        :param thread:       not supported
        :return: breakpoint number
        """
        if not hardware:
            self.log.warning('Software breakpoints are not supported, falling back to hardware')
        if regex:
            self.log.warning('Regex breakpoints are not supported, ignoring regex')
        if condition is not None:
            self.log.warning('Conditional breakpoints are not supported, ignoring condition')
        if thread:
            self.log.warning('Thread-specific breakpoints are not supported, ignoring thread')
        # TODO line <-> addr
        bkptno = len(self._breakpoints)
        hook = self.uc.hook_add(unicorn.UC_HOOK_CODE, self._breakpoint_hook, begin=line,
                                end=line, user_data=bkptno)
        self._breakpoints.append(UnicornBreakpoint(hooks=[hook], temporary=temporary,
                                                   ignore_count=ignore_count))
        return bkptno

    def set_watchpoint(self, variable, write=True, read=False):
        """Insert a watchpoint.
        This is currently NOT WORKING because of a bug in Unicorn.
        See https://github.com/unicorn-engine/unicorn/issues/972 for further details.

        :param variable: address to watch
        :param write:    whether to watch writes
        :param read:     whether to watch reads
        :return: watchpoint number
        """
        # TODO variable <-> addr
        bkptno = len(self._breakpoints)
        hooks = []
        if write is True:
            hooks.append(self.uc.hook_add(unicorn.UC_HOOK_MEM_WRITE, self._watchpoint_hook,
                                          begin=variable, end=variable, user_data=bkptno))
        if read is True:
            hooks.append(self.uc.hook_add(unicorn.UC_HOOK_MEM_READ, self._watchpoint_hook,
                                          begin=variable, end=variable, user_data=bkptno))
        self._breakpoints.append(UnicornBreakpoint(hooks=hooks))
        return bkptno

    def remove_breakpoint(self, bkptno):
        """Remove a breakpoint or watchpoint.

        :param bkptno: breakpoint/watchpoint number
        """
        for hook in self._breakpoints[bkptno].hooks:
            self.uc.hook_del(hook)
        self._breakpoints[bkptno] = None

    # Memory protocol

    def read_memory(self, address, wordsize, num_words=1, raw=False):
        """Read memory.

        :param address:   the address to read from
        :param wordsize:  the size of a read word (1, 2, 4 or 8)
        :param num_words: the amount of words to read
        :param raw:       whether the read memory should be returned unprocessed
        :return: the read memory
        :rtype:  int if num_words == 1 and raw == False,
                 list of int if num_words > 1 and raw == False,
                 str or byte if raw == True
        """
        raw_mem = self.uc.mem_read(address, wordsize * num_words)
        if raw:
            return raw_mem

        # TODO: endianness support
        num2fmt = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}
        fmt = '<{}{}'.format(num_words, num2fmt[wordsize])
        mem = struct.unpack(fmt, raw_mem)
        return mem[0] if num_words == 1 else mem

    def write_memory(self, address, wordsize, val, num_words=1, raw=False):
        """Write memory.

        :param address:   the address to write to
        :param wordsize:  size of a written word (1, 2, 4 or 8)
        :param val:       the value to write
        :type val:        int if num_words == 1 and raw == False,
                          list of int if num_words > 1 and raw == False,
                          str or byte if raw == True
        :param num_words: the amount of words to write
        :param raw:       whether to write in raw mode
        :return: True on success, False otherwise
        """
        if raw:
            raw_mem = val
        else:
            # TODO: endianness support
            num2fmt = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}
            fmt = '<{}{}'.format(num_words, num2fmt[wordsize])
            if num_words == 1:
                raw_mem = struct.pack(fmt, val)
            else:
                raw_mem = struct.pack(fmt, *val)

        try:
            self.uc.mem_write(address, raw_mem)
            return True
        except unicorn.UcError:
            self.log.debug('Failed memory write @ 0x{:x}'.format(address))
            return False

    # Register protocol

    def write_register(self, reg, value):
        """Write a register.

        :param reg:   name of the register to write
        :param value: value to write
        """
        self.uc.reg_write(self.arch.unicorn_registers[reg], value)

    def read_register(self, reg):
        """Read a register.

        :param reg: name of the register to read
        :return: read value
        """
        return self.uc.reg_read(self.arch.unicorn_registers[reg])

    # Remote memory protocol

    def send_response(self, id, value, success):
        """Handle a remote memory response.

        :param id:      the request ID
        :param value:   read value, if it was a read request
        :param success: True if the request was successful, False otherwise
        :return: True if the response was handled successfully, False otherwise
        """
        self._rmp_queue.put((value, success))
        return True

    # ---

    def _forward_hook(self, uc, access, address, size, value, user_data):
        """Unicorn hook for memory forwarding."""
        pc = self.read_register(self.arch.pc_name)
        if access == unicorn.UC_MEM_READ or access == unicorn.UC_MEM_FETCH:
            msg = RemoteMemoryReadMessage(self._origin, 0, pc, address, size)
            write_back = True
        elif access == unicorn.UC_MEM_WRITE:
            msg = RemoteMemoryWriteMessage(self._origin, 0, pc, address, value, size)
            write_back = False
        else:
            raise ValueError('Forward hook with unknown access {}'.format(access))

        self._avatar_queue.put(msg)
        value, success = self._rmp_queue.get()
        if not success:
            self.log.debug('Remote memory request returned 0x{:x}'.format(value))
        elif write_back and not self.write_memory(address, size, value):
            self.log.debug('Failed to write back remote memory')

    def _breakpoint_hook(self, uc, address, size, bkptno):
        """Unicorn hook for breakpoints."""
        if bkptno in self.pending_bp:
            return
        bp = self._breakpoints[bkptno]
        if bp.ignore_count > 0:
            bp.ignore_count -= 1
            return

        self.pending_bp.add(bkptno)
        self._worker_queue.put(UnicornWorkerBreakpointMessage(bkptno, address))
        self.uc.emu_stop()

        if bp.temporary:
            self.remove_breakpoint(bkptno)

    def _watchpoint_hook(self, uc, access, address, size, value, bkptno):
        """Unicorn hook for watchpoints."""
        if bkptno in self.pending_bp:
            return
        self.pending_bp.add(bkptno)
        self.stop()

    def _worker_emu_start(self, single_step=False):
        """Start the emulation inside the worker."""
        self._worker_queue.put(UnicornWorkerEmuStartMessage(single_step))

    def _worker_emu_stop(self):
        """Stop the emulation inside the worker."""
        self._worker_queue.put(UnicornWorkerUpdateStateMessage(TargetStates.STOPPED))
        self.uc.emu_stop()


class UnicornWorker(Thread):
    """Worker class for the Unicorn protocol."""

    def __init__(self, origin, protocol, uc, worker_queue, avatar_queue):
        self._origin = origin
        self._protocol = protocol
        self._uc = uc
        self._worker_queue = worker_queue
        self._avatar_queue = avatar_queue
        super(UnicornWorker, self).__init__()

    def run(self):
        while True:
            msg = self._worker_queue.get()
            if msg is None:
                break
            if isinstance(msg, UnicornWorkerEmuStartMessage):
                self._avatar_queue.put(UpdateStateMessage(self._origin, TargetStates.RUNNING))
                if self._protocol.pending_bp:
                    # Single-step over pending breakpoints, the hook will ignore them
                    old_pending_bp = self._protocol.pending_bp.copy()
                    self._uc.emu_start(self._get_next_pc(), self._EMU_END_ADDRESS, count=1)
                    if self._protocol.pending_bp == old_pending_bp:
                        # We did not hit another breakpoint during the step, empty the pending set
                        self._protocol.pending_bp.clear()
                    if msg.single_step:
                        # We already stepped: done
                        continue
                if not self._protocol.pending_bp:
                    # Either there was no pending breakpoint, or we didn't hit
                    # another breakpoint while single stepping: keep running
                    count = 1 if msg.single_step else 0
                    self._uc.emu_start(self._get_next_pc(), self._EMU_END_ADDRESS, count=count)
            elif isinstance(msg, UnicornWorkerUpdateStateMessage):
                self._avatar_queue.put(UpdateStateMessage(self._origin, msg.state))
            elif isinstance(msg, UnicornWorkerBreakpointMessage):
                # When stopping from a hook, Unicorn resets the PC to the beginning of the basic
                # block that contains the instruction that triggered the hook.
                # The register state, however, isn't rolled back.
                # As a workaround, we set the PC here to the breakpoint address.
                # See https://github.com/unicorn-engine/unicorn/issues/969
                self._protocol.write_register(self._protocol.arch.pc_name, msg.address)
                self._avatar_queue.put(BreakpointHitMessage(self._origin, msg.bkptno, msg.address))
            else:
                raise Exception('Unknown message in Unicorn worker queue: {}'.format(msg))

    def _get_next_pc(self):
        """Get the PC to start emulation at."""
        pc = self._protocol.read_register(self._protocol.arch.pc_name)
        return self._fixup_thumb_pc(pc)

    def _fixup_thumb_pc(self, pc):
        """Fix the PC for emu_start to take ARM Thumb mode into account."""
        # If the arch mode is UC_MODE_THUMB, force Thumb.
        # Otherwise, check Thumb bit in CPSR.
        if self._protocol.arch.unicorn_arch == unicorn.UC_ARCH_ARM and \
                (self._protocol.arch.unicorn_mode == unicorn.UC_MODE_THUMB or
                 self._protocol.read_register(self._protocol.arch.sr_name) & 0x20):
            pc |= 1
        return pc

    _EMU_END_ADDRESS = 0  # TODO: 0 could be a valid address



from types import MethodType
from threading import Event
from enum import Enum
import logging

from avatar2.watchmen import AFTER, BEFORE, watch
from avatar2.archs import ARM, ARMBE
from avatar2 import TargetStates

# From the ARM CoreSight Architecture Specification v2.0
component_id_registers = {
    'CIDR0': 0xff0,
    'CIDR1': 0xff4,
    'CIDR2': 0xff8,
    'CIDR3': 0xffc
}

device_type_register_offset = 0xFCC

device_types = {
    0x0: ("Miscellaneus", {
        0x0: "Other, unidentified",
        0x1: "Validation component"}),
    0x1: ("Trace Sink", {
        0x0: "Other",
        0x1: "Trace port, for example TPIU",
        0x2: "Buffer, for example ETB",
        0x3: "Basic trace router"}),
    0x2: ("Trace Link", {
        0x0: "Other",
        0x1: "Trace funnel, Router",
        0x2: "Filter",
        0x3: "FIFO, Large Buffer"}),
    0x3: ("Trace Source", {
        0x0: "Other",
        0x1: "Associated with a processor core",
        0x2: "Associated with a DSP",
        0x3: "Associated with a Data Engine or Coprocessor",
        0x4: "Associated with a Bus, stimulus derived from bus activity",
        0x6: "Associated with software, stimulus derived from software activity"}),
    0x4: ("Debug Control", {
        0x0: "Other",
        0x1: "Trigger Matrix, for example ECT",
        0x2: "Debug Authentication Module",
        0x3: "Power requestor"}),
    0x5: ("Debug Logic", {
        0x0: "Other",
        0x1: "Processor core",
        0x2: "DSP",
        0x3: "Data Engine or Coprocessor",
        0x4: "Bus, stimulus derived from bus activity",
        0x5: "Memory, tightly coupled device such as Built In Self Test (BIST)"}),
    0x6: ("Performance monitor", {
        0x0: "Other",
        0x1: "Associated with a processor",
        0x2: "Associated with a DSP",
        0x3: "Associated with a Data Engine or Coprocessor",
        0x4: "Associated with a bus, stimulus derived from bus activity",
        0x5: ("Associated with a memory management unit that"
              + "conforms to the ARM System MMU Architecture")})
}


def read_id_registers(target, base_address):
    return dict((name, target.rm(base_address + offset, size=4)) for
                name, offset in component_id_registers.items())


def parse_coresight_type(device_type):
    maj_value = device_type & 0x7
    sub_value = (device_type >> 4) & 0x7
    return (device_types[maj_value][0],
            device_types[maj_value][1][sub_value])


def find_coresight_magic_value(avatar, message):
    if message.value != 0xC5ACCE55 or message.size != 4:
        return

    log = logging.getLogger('avatar')
    base_address = message.address & 0xfffff000
    log.info("Detected possible ARM Coresight Debug component" +
             "at address 0x%x" % base_address)
    target = avatar.get_memory_range(base_address).forwarded_to

    regs_value = read_id_registers(target, base_address)
    if (regs_value['CIDR0'] != 0xD
        or regs_value['CIDR1'] != 0x90
        or regs_value['CIDR2'] != 0x5
        or regs_value['CIDR3'] != 0xB1):
        log.info("Detecting Coresight component: False positive")
        return

    device_type = target.rm(base_address + device_type_register_offset, size=4)
    log.info("Device type 0x%x" % (device_type))
    (maj, sub) = parse_coresight_type(device_type)
    log.info("Detected Coresight component. Major type: %s. Sub type: %s"
             % (maj, sub))


def load_plugin(avatar):
    if avatar.arch not in [ARM, ARMBE]:
        log = logging.getLogger('avatar')
        log.warning("Coresight components are part of the ARM architecture.")

    avatar.watchmen.add_watchman('RemoteMemoryWrite', when=BEFORE,
                                 callback=find_coresight_magic_value,
                                 is_async=True)



from configparser import ConfigParser
from distutils.dir_util import mkpath
from os.path import expanduser, realpath, dirname, exists
from distutils.spawn import find_executable as find
from collections import OrderedDict

CONFIG_FILE = expanduser('~/.avatar2/settings.cfg')

# Constant names for the different targets
OPENOCD = 'openocd'
QEMU = 'avatar-qemu'
PANDA = 'panda'
GDB_MULTI = 'gdb (multiarch)'
GDB_X86 = 'gdb (x86)' 


TARGETS = OrderedDict(
    [
    (OPENOCD, { 'git': 'https://git.code.sf.net/p/openocd/code',
               'configure': '',
               'make': '',
               'git_apt_deps': 'libtool pkg-config autoconf automake texinfo',
               'rel_path': 'src/openocd',
               'install_cmd': ['./bootstrap','./configure','make'],
               'apt_name': 'openocd'
             }),
    (QEMU, {  'git': 'https://github.com/avatartwo/avatar-qemu',
             'configure': '--disable-sdl --target-list=arm-softmmu',
             'make': '',
             'git_apt_deps': 'libglib2.0-dev libfdt-dev libpixman-1-dev zlib1g-dev',
             'rel_path': 'arm-softmmu/qemu-system-arm',
             'install_cmd': ['git submodule update --init dtc',
                             './configure', 'make']
          }),
    (PANDA, {'git': 'https://github.com/panda-re/panda',
             'configure': '--disable-sdl --target-list=arm-softmmu',
             'make': '',
             'git_apt_deps': '',
             'rel_path': 'arm-softmmu/qemu-system-arm',
             'install_cmd': ['git submodule update --init dtc',
                             './configure', 'make'],
             'install_script': 'panda/scripts/install_ubuntu.sh'
           }),
    (GDB_X86, { 'apt_name': 'gdb' }),
    (GDB_MULTI, { 'apt_name': 'gdb-multiarch'})
    ]
)


class AvatarConfig(ConfigParser):
    

    def __init__(self):
        super(AvatarConfig, self).__init__()
        self.config_file = realpath(expanduser(CONFIG_FILE))
        self.config_path = dirname(self.config_file)

        mkpath(expanduser(self.config_path)) # create config dir if neccessary

        # Create a default config if there's no config file yet
        if self.read(expanduser(CONFIG_FILE)) == []:
            self.add_section('DIST')
            self.add_section('TARGETS')

            has_apt = 'True' if find('apt-get') else 'False'
            self.set('DIST', 'has_apt', has_apt)
            self.set('DIST', 'default_install_path', self.config_path+'/')

            for t_name, t_dict in TARGETS.items():
                path = t_dict.get('sys_name', t_dict.get('apt_name', 'None'))
                full_path = find(path) or 'None'
                self.set('TARGETS', t_name, full_path) 


    def write_config(self):
        with open(expanduser(CONFIG_FILE), 'w+') as cfgfile:
            self.write(cfgfile)


    def get_target_path(self, target):
        if self.has_section('TARGETS'):
            target_path = self.get('TARGETS', target)
            return None if target_path == 'None' else target_path



from types import MethodType

from capstone import *


def forward_instructions(self, from_target, to_target,
                         memory_region, instructions=None,
                         read_from_file=True):
    if instructions is None:
        instructions = self.arch.unemulated_instructions

    if memory_region.forwarded:
        raise Exception("Cannot forward instructions from forwarded" +
                        " memory region")

    if read_from_file and not memory_region.file:
        raise Exception("No file specified for this memory region")

    if read_from_file:
        with open(memory_region.file, 'rb') as f:
            content = f.read()
    else:
        content = from_target.rm(memory_region.address, memory_region.size)

    md = Cs(self.arch.capstone_arch, self.arch.capstone_mode)
    for (addr, size, op, _) in md.disasm_lite(content, memory_region.address):
        if op in instructions:
            self.log.debug("%s instruction found at %x. " +
                           "Adding transition.")
            self.add_transition(addr, from_target, to_target,
                                sync_regs=True)
            self.add_transition(addr + size, to_target, from_target,
                                sync_regs=True)


def load_plugin(avatar):
    if 'orchestrator' not in avatar.loaded_plugins:
        avatar.load_plugin('orchestrator')
    avatar.forward_instructions = MethodType(forward_instructions, avatar)





class AvatarMessage(object):
    def __init__(self, origin):
        self.origin = origin

    def __str__(self):
        if self.origin:
            return "%s from %s" % (self.__class__.__name__, self.origin.name)
        else:
            return "%s from unkown origin" % self.__class__.__name__


class UpdateStateMessage(AvatarMessage):
    def __init__(self, origin, new_state):
        super(UpdateStateMessage, self).__init__(origin)
        self.state = new_state


class BreakpointHitMessage(UpdateStateMessage):
    def __init__(self, origin, breakpoint_number, address):
        super(BreakpointHitMessage, self).__init__(origin, TargetStates.BREAKPOINT)
        self.breakpoint_number = breakpoint_number
        self.address = address

class SyscallCatchedMessage(BreakpointHitMessage):
    def __init__(self, origin, breakpoint_number, address, type='entry'):
        super(self.__class__, self).__init__(origin, breakpoint_number, address)
        self.type = type


class RemoteMemoryReadMessage(AvatarMessage):
    def __init__(self, origin, id, pc, address, size, dst=None):
        super(self.__class__, self).__init__(origin)
        self.id = id
        self.pc = pc
        self.address = address
        self.size = size
        self.dst = dst
        self.num_words = 1
        self.raw = False

class RemoteMemoryWriteMessage(AvatarMessage):
    def __init__(self, origin, id, pc, address, value, size, dst=None):
        super(self.__class__, self).__init__(origin)
        self.id = id
        self.pc = pc
        self.address = address
        self.value = value
        self.size = size
        self.dst = dst

class RemoteInterruptEnterMessage(AvatarMessage):
    def __init__(self, origin, id, interrupt_num):
        super(self.__class__, self).__init__(origin)
        self.id = id
        self.interrupt_num = interrupt_num

class RemoteInterruptExitMessage(AvatarMessage):
    def __init__(self, origin, id, transition_type, interrupt_num):
        super(self.__class__, self).__init__(origin)
        self.id = id
        self.transition_type = transition_type
        self.interrupt_num = interrupt_num


from .targets.target import TargetStates



import sys
import subprocess
import telnetlib
import logging
import distutils
from codecs import encode
import binascii
from threading import Thread, Lock, Event
from struct import pack, unpack
from time import sleep
import re
from os.path import abspath
if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

from avatar2.targets import TargetStates
from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage

END_OF_MSG = u'\x1a'


class OpenOCDProtocol(Thread):
    """
    This class implements the openocd protocol.

    :param openocd_script:     The openocd scripts to be executed.
    :type openocd_script:      str or list
    :param openocd_executable: The executable
    :param additional_args:    Additional arguments delivered to openocd.
    :type  additional_args:    list
    :param tcl_port:        the port used for the telnet connection
    :param gdb_port:           the port used for openocds gdb-server
    """

    def __init__(self, avatar, origin, openocd_script, openocd_executable="openocd",
                 additional_args=[], host='127.0.0.1', tcl_port=6666, gdb_port=3333,
                 output_directory='/tmp', debug=False):
        """
        OpenOCD machine interface protocol
        :param avatar: The Avatar instance
        :param origin: The Target this protocol belongs to
        :param openocd_script: The OpenOCD script for starting and configuring the taret.  Should not call 'init'
        :param openocd_executable: The path to the OpenOCD you want to use
        :param additional_args: Additional args to openocd
        :param host: THe host of the running OpenOCD.  Probably localhost
        :param tcl_port: The port for OpenOCD's TCL machine interface.  Should be 6666
        :param gdb_port: The GDB port for OpenOCD's GDB interface
        :param output_directory: The directory where logfiles should go
        :param debug: Enable openocd debug output
        """
        if isinstance(openocd_script, str):
            self.openocd_files = [openocd_script]
        elif isinstance(openocd_script, list):
            self.openocd_files = openocd_script
        else:
            raise TypeError("Wrong type for OpenOCD configuration files")
        self.log = logging.getLogger('%s.%s' % (origin.log.name, self.__class__.__name__)) if origin else \
                                    logging.getLogger(self.__class__.__name__)
        self._tcl_port = tcl_port
        self._gdb_port = gdb_port
        self._host = host
        self.in_queue = queue.Queue()
        self.out_queue = queue.Queue()
        self.trace_queue = queue.Queue()
        self.trace_enabled = Event()
        self.avatar = avatar
        self.telnet = None
        self._close = Event()
        self.buf = u""
        self.cmd_lock = Lock()
        self._origin = origin

        self.output_directory = output_directory
        executable_path = distutils.spawn.find_executable(openocd_executable)
        self._cmd_line = [executable_path]
        if debug is True:
            self._cmd_line += ['--debug']

        self._cmd_line += [e for l
                           in [['-f', abspath(f)] for f in self.openocd_files]
                           for e in l]
        self._cmd_line += ['--command', 'tcl_port %d' % self._tcl_port,
                          '--command', 'gdb_port %d' % self._gdb_port]
        self._cmd_line += additional_args

        self._openocd = None

        with open("%s/openocd_out.txt" % output_directory, "wb") as out, \
                open("%s/openocd_err.txt" % output_directory, "wb") as err:
            self.log.debug("Starting OpenOCD with command line: %s" % (" ".join(self._cmd_line)))
            self._openocd = subprocess.Popen(self._cmd_line,
                                             stdout=out, stderr=err)#, shell=True)
        Thread.__init__(self)
        self.daemon = True


    def connect(self):
        """
        Connects to OpenOCDs TCL Server for all subsequent communication
        returns: True on success, else False
        """
        sleep(1)
        
        if self._openocd.poll() is not None:
            raise RuntimeError(("Openocd errored! Please check "
                                "%s/openocd_err.txt for details" %
                                self.output_directory))


        self.log.debug("Connecting to OpenOCD on %s:%s" % (self._host, self._tcl_port))
        try:
            self.telnet = telnetlib.Telnet(self._host, self._tcl_port)
            # mic check
            self.telnet.write("ocd_echo\x1a".encode('ascii'))
            self.log.debug("Connected to OpenOCD.  Saying hello...")
            stuff = self.telnet.read_until('\x1a'.encode('ascii'))
            self.log.debug("Got a hello back.  Starting background thread...")
            self.start()
            # One last thing, enable target notifications!
            self.log.debug("Enabling async target notifications...")
            self.execute_command("tcl_notifications on")
            return True
        except:
            self.log.exception("Error connecting to OpenOCD TCL port %d" % self._tcl_port)
            return False

    def enable_trace(self):
        """
        Enables OpenOCD's "target tracing" support.
        Requires a version of OpenOCD >= 0.10.0-dev
        (the absolute latest one is probably a good idea)
        This will log all the trace data to the special `trace_queue` member.
        Something else (like the CoreSightProtocol) should figure out what to do with that.
        :return:
        """
        self.log.debug("Enabling tracing...")
        resp = self.execute_command("ocd_tcl_trace on")
        if 'is enabled' in resp:
            self.trace_enabled.set()
            return True
        else:
            self.log.warning("Could not enable target tracing! Is your OpenOCD old?")
            return False

    def handle_target_notification(self, str):

        # Check if we should handle notifications
        if self != self._origin.protocols.execution:
            return

        # mode halt | run | init
        reset_re = re.compile("type target_reset mode (\w+)")
        # Trace data (from TPIU) Hex-encoded.
        trace_re = re.compile("type target_trace data ([0-9a-f]+)")
        # State change
        state_re = re.compile("type target_state state (\S+)")
        # Generic event
        event_re = re.compile("type target_event event (\S+)")

        mreset = reset_re.match(str)
        mtrace = trace_re.match(str)
        mstate = state_re.match(str)
        mevent = event_re.match(str)

        if mreset:
            reset_type = mreset.group(1)
            self.log.debug("Got reset event type %s" % reset_type)
            if reset_type == "halt":
                avatar_msg = UpdateStateMessage(self._origin, TargetStates.STOPPED)
                self.avatar.fast_queue.put(avatar_msg)
        elif mtrace:
            # DOn't log anything here.  If we do, our IO will be exhausted by the sheer volume of trace packets.
            self.trace_queue.put(str)
        elif mstate:
            # Do a state update!
            state = mstate.group(1)
            if "halted" in state:
                self.log.debug("Target has halted")
                avatar_msg = UpdateStateMessage(self._origin, TargetStates.STOPPED)
                self.avatar.fast_queue.put(avatar_msg)
            elif "running" in state:
                self.log.debug("Target is now running")
                avatar_msg = UpdateStateMessage(self._origin, TargetStates.RUNNING)
                self.avatar.fast_queue.put(avatar_msg)
            else:
                self.log.warning("Weird target state %s" % state)
        elif mevent:
            #TODO handle these
            event = mevent.group(1)
            self.log.debug("Target event: %s " % event)
            # TODO handle these
            if event == 'halted':
                avatar_msg = UpdateStateMessage(self._origin, TargetStates.STOPPED)
                self.avatar.fast_queue.put(avatar_msg)
            elif event == 'resumed':
                avatar_msg = UpdateStateMessage(self._origin, TargetStates.RUNNING)
                self.avatar.fast_queue.put(avatar_msg)

        else:
            self.log.warning("Unhandled event message %s" % str)


    def reset(self):
        """
        Resets the target
        returns: True on success, else False
        """
        self.log.debug("Resetting target")
        resp = self.execute_command('reset halt')
        if not 'Not halted' in str(resp):
            self.log.debug("Target reset complete")
            return True
        else:
            self.log.error('Failed to reset the target with OpenOCD')
            return False

    def shutdown(self):
        """
        Shuts down OpenOCD
        returns: True on success, else False
        """
        #self.execute_command('ocd_shutdown')
        self._close.set()
        if self.telnet:
            self.telnet.close()
        # Fix
        if self._openocd is not None:
            self._openocd.terminate()
            self._openocd = None

    def execute_command(self, cmd):
        try:
            self.cmd_lock.acquire()
            self.in_queue.put(cmd)
            ret = self.out_queue.get()
            if "FAILED" in ret:
                raise RuntimeError("Command '%s' failed!" % cmd)
            return ret
        except:
            raise
        finally:
            self.cmd_lock.release()

    def run(self):
        try:
            cmd = None
            self.log.debug("Starting OpenOCDSocketListener")
            while not self.avatar._close.is_set() and not self._close.is_set():
                if not self.in_queue.empty():
                    cmd = self.in_queue.get()
                    self.log.debug("Executing command %s" % cmd)
                    self.telnet.write((cmd + END_OF_MSG).encode('ascii'))
                try:
                    line = self.read_response()
                except EOFError:
                    self.log.warning("OpenOCD Connection closed!")
                    self.shutdown()
                    break
                if line is not None:
                    #print line
                    line = line.rstrip(END_OF_MSG)
                    # This is async target notification data.  Don't return it normally
                    if line.startswith("type"):
                        self.handle_target_notification(line)
                    # This is an error
                    elif "Error" in line:
                        self.log.error(line)
                        if cmd:
                            # tell the caller we failed
                            self.out_queue.put("FAILED")
                    else:
                        if not cmd:
                            # We didn't ask for it.  Just debug it
                            self.log.debug(line)
                        else:
                            self.log.debug("response --> " +  line)
                            self.out_queue.put(line)
                            cmd = None
                sleep(.001) # Have a heart. Give other threads a chance
        except Exception as e:
            self.log.exception("OpenOCD Background thread died with an exception")
        self.log.debug("OpenOCD Background thread exiting")

    def read_response(self):
        self.buf += self.telnet.read_eager().decode('ascii')
        #if buf is not '':
            #print(self.buf)
        if END_OF_MSG in self.buf:
            resp, self.buf = self.buf.split(END_OF_MSG, 1)
            return resp
        return None

    ### The Memory Protocol starts here

    def write_memory(self, address, wordsize, val, num_words=1, raw=False):
        """Writes memory

        :param address:   Address to write to
        :param wordsize:  the size of the write (1, 2, 4 or 8)
        :param val:       the written value
        :type val:        int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """
        #print "nucleo.write_memory(%s, %s, %s, %s, %s)" % (repr(address), repr(wordsize), repr(val), repr(num_words), repr(raw))
        if isinstance(val, str) and len(val) != num_words:
            self.log.debug("Setting num_words = %d" % (len(val) / wordsize))
            num_words = len(val) / wordsize
        for i in range(0, num_words, wordsize):
            if raw:
                write_val = '0x' + encode(val[i:i+wordsize], 'hex_codec').decode('ascii')
            elif isinstance(val, int) or isinstance(val, long):
                write_val = hex(val).rstrip("L")
            else:
                # A list of ints
                write_val = hex(val[i]).rstrip("L")
            write_addr = hex(address + i).rstrip("L")
            if wordsize == 1:
                self.execute_command('mwb %s %s' % (write_addr, write_val))
            elif wordsize == 2:
                self.execute_command('mwh %s %s' % (write_addr, write_val))
            else:
                self.execute_command('mww %s %s' % (write_addr, write_val))

        return True

    def read_memory(self, address, wordsize=4, num_words=1, raw=False):
        """reads memory

        :param address:   Address to write to
        :param wordsize:  the size of a read word (1, 2, 4 or 8)
        :param num_words: the amount of read words
        :param raw:       Whether the read memory should be returned unprocessed
        :return:          The read memory
        """
        num2fmt = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}
        raw_mem = b''
        words = []
        for i in range(0, num_words, wordsize):
            read_addr = hex(address + i).rstrip('L')
            if wordsize == 1:
                resp = self.execute_command('mrb %s' % read_addr)
            elif wordsize == 2:
                resp = self.execute_command("mrh %s" % read_addr)
            else:
                resp = self.execute_command('mrw %s' % read_addr)
            if resp:
                val = int(resp)
                raw_mem += binascii.unhexlify(hex(val)[2:].zfill(wordsize * 2))
            else:
                self.log.error("Could not read from address %s" % read_addr)
                return None
        # OCD flips the endianness
        raw_mem = raw_mem[::-1]
        if raw:
            self.log.debug("Read %s from %#08x" % (repr(raw), address))
            return raw_mem
        else:
            # Todo: Endianness support
            fmt = '<%d%s' % (num_words, num2fmt[wordsize])
            mem = list(unpack(fmt, raw_mem))
            if num_words == 1:
                return mem[0]
            else:
                return mem

    ### The register protocol starts here

    def read_register(self, reg):

        try:
            resp = self.execute_command("ocd_reg %s" % reg)
            val = int(resp.split(":")[1].strip(), 16)
            return val
        except:
            self.log.exception("Failed to read from register " + repr(reg))
            return False

    def write_register(self, reg, value):
        """Set one register on the target
        :returns: True on success"""
        try:
            self.execute_command("ocd_reg %s %s" % (reg, hex(value)))
            return True
        except:
            self.log.exception(("Error writing register %s" % reg))
            return False

    def cont(self):
        """Continues the execution of the target
        :returns: True on success"""
        try:
            resp = self.execute_command("resume")
        except:
            self.log.exception("Error halting target")
            return False
        return True

    def stop(self):
        """Stops execution of the target
        :returns: True on success"""
        try:
            resp = self.execute_command("halt")
        except:
            self.log.exception("Error halting target")
            return False
        return True

    def step(self):
        """Step one instruction on the target
        :returns: True on success"""
        try:
            resp = self.execute_command("step")
            return True
        except:
            self.log.exception("Failed to step the target")

    def set_breakpoint(self, line,
                       hardware=False,
                       temporary=False,
                       regex=False,
                       condition=None,
                       ignore_count=0,
                       thread=0):
        """Inserts a breakpoint
        :param str line: the thing to break at.  An address.
        :param bool hardware: Hardware breakpoint
        :param bool temporary:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with specified condition
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        :returns:             The number of the breakpoint
        """
        cmd = ["bp"]
        if regex:
            raise ValueError("OpenOCD doesn't support regex breakpoints!")
        if condition:
            raise ValueError("OpenOCD doesn't support conditional breakpoints!")
        if ignore_count:
            raise ValueError("OpenOCD doesn't support ignore counts")
        if thread:
            raise ValueError("OpenOCD doesn't support thread options!")

        if isinstance(line, int):
            cmd.append("%#08x" % line)
        else:
            cmd.append(str(line))
        cmd.append("2") # TODO: This isn't platform-independent, but i have no idea what it does
        if hardware:
            cmd.append("hw")
        try:
            resp = self.execute_command(" ".join(cmd))
            self.log.debug("Breakpoint set")
            return True
        except:
            self.log.exception("Error setting breakpoint")
            return False

    def set_watchpoint(self, variable, write=True, read=False):
        cmd = ["wp"]

        if isinstance(variable, int):
            cmd.append("%#08x" % variable)
        else:
            cmd.append(str(variable))
        cmd.append("2") # TODO FIXME
        if read and write:
            cmd.append("a")
        elif read:
            cmd.append("r")
        elif write:
            cmd.append("w")
        else:
            raise ValueError("At least one read and write must be True")
        try:
            resp = self.execute_command(" ".join(cmd))
            return True
        except:
            self.log.exception("Error setting watchpoint")
            return False

    def remove_breakpoint(self, bkpt):
        """Deletes a breakpoint"""
        cmd = ['rbp']
        if isinstance(bkpt, int):
            cmd.append("%#08x" % bkpt)
        else:
            cmd.append(str(bkpt))
        try:
            self.execute_command(" ".join(cmd))
            return True
        except:
            self.log.exception("Error removing breakpoint")




import sys

if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

import atexit
import tempfile
import intervaltree
import logging
import signal
import json
import time

from copy import deepcopy
from os import path, makedirs
from threading import Thread, Event

from .archs.arm import ARM
from .memory_range import MemoryRange
from .message import *
from .peripherals import AvatarPeripheral
from .targets.target import TargetStates  # TargetStates
from .watchmen import watch, Watchmen


class Avatar(Thread):
    """The Avatar-object is the main interface of avatar.
    Here we can set the different targets, and more

    :ivar arch:     architecture of all targets
    :ivar endness:  used endianness


    """

    def __init__(
        self, arch=ARM, cpu_model=None, output_directory=None, log_to_stdout=True, configure_logging=True
    ):
        super(Avatar, self).__init__()

        self.shutdowned = False
        try:
            signal.signal(signal.SIGINT, self.sigint_wrapper)
            self.sigint_handler = self.shutdown
        except ValueError:
            # Cannot register SIGINT handler: we are not in main thread. Do not care about it.
            pass
        atexit.register(self.shutdown)

        self.watchmen = Watchmen(self)
        self.arch = arch
        self.arch.init(self)
        self.targets = {}
        self.status = {}
        self.memory_ranges = intervaltree.IntervalTree()
        self.loaded_plugins = []
        self.cpu_model = cpu_model

        if self.cpu_model is None and hasattr(self.arch, "cpu_model"):
            self.cpu_model = self.arch.cpu_model
        # Setup output-dir and logging
        self.output_directory = (
            tempfile.mkdtemp(suffix="_avatar")
            if output_directory is None
            else output_directory
        )
        if not path.exists(self.output_directory):
            makedirs(self.output_directory)

        self.log = logging.getLogger("avatar")
        
        if configure_logging:
            format = "%(asctime)s | %(name)s.%(levelname)s | %(message)s"

            logfile = "%s/avatar.log" % self.output_directory
            logging.basicConfig(
                filename=logfile,
                level=logging.INFO,
                format=format,
            )

            if log_to_stdout is True:
                handler = logging.StreamHandler(sys.stdout)
                handler.setFormatter(logging.Formatter(format))
                root = logging.getLogger()
                root.addHandler(handler)

        self.log.info(
            "Initialized Avatar. Output directory is %s" % self.output_directory
        )

        # Setup the avatarqueues and register default handler
        self._close = Event()
        self.queue = queue.Queue()
        self.fast_queue = queue.Queue()
        self.fast_queue_listener = AvatarFastQueueProcessor(self)
        self.message_handlers = {
            SyscallCatchedMessage: self._handle_syscall_catched_message,
            BreakpointHitMessage: self._handle_breakpoint_hit_message,
            UpdateStateMessage: self._handle_update_state_message,
            RemoteMemoryReadMessage: self._handle_remote_memory_read_message,
            RemoteMemoryWriteMessage: self._handle_remote_memory_write_message,
        }
        self.daemon = True
        self.start()

    def load_config(self, file_name=None):
        """
        Populates the avatar object with targets and ranges saved in a config
        json. Note that some changes on the targets may be lost, for instance
        custom protocol configuration.
        :ivar file_name: (Absolute) path to the config file
        """
        if file_name is None:
            file_name = "%s/conf.json" % self.output_directory
        with open(file_name, "r") as config_file:
            config = json.load(config_file)

        for t in config.pop("targets", []):
            module = __import__(t.pop("module"))
            klass = getattr(module, t.pop("type"))
            self.add_target(klass, **t)

        for mr in config.pop("memory_mapping", []):
            # resolve forwarded_to to the target objects
            tname = mr.get("forwarded_to")
            if tname is not None:
                if not tname in self.targets:
                    raise Exception(
                        (
                            "Requested target %s not found in config. "
                            "Aborting." % tname
                        )
                    )
                mr["forwarded_to"] = self.targets[tname]
            # TODO handle emulate
            self.add_memory_range(mr.pop("address"), mr.pop("size"), **mr)
        for k, v in config.items():
            setattr(self, k, v)

    def generate_config(self):
        """
        Generates a configuration dictionary for storage based on the currently
        defined targets and memory ranges
        """
        conf_dict = {}
        if self.cpu_model is not None:
            conf_dict["cpu_model"] = self.cpu_model
        conf_dict["memory_mapping"] = []
        for mr in self.memory_ranges:
            conf_dict["memory_mapping"].append(mr.data.dictify())

        conf_dict["targets"] = []
        for t in self.targets.values():
            conf_dict["targets"].append(t.dictify())

        return conf_dict

    def save_config(self, file_name=None, config=None):
        if file_name is None:
            file_name = "%s/conf.json" % self.output_directory
        conf_dict = self.generate_config() if config is None else config
        with open(file_name, "w") as conf_file:
            json.dump(conf_dict, conf_file)

    def shutdown(self):
        """
        Shuts down all targets and Avatar. Should be called at end of script
        in order to cleanly exit all spawned processes and threads
        """
        if self.shutdowned is True:
            return
        for t in self.targets.values():
            t.shutdown()
        for range in self.memory_ranges:
            if isinstance(range.data.forwarded_to, AvatarPeripheral):
                range.data.forwarded_to.shutdown()
        self.shutdowned = True

    def sigint_wrapper(self, signal, frame):
        self.log.info("Avatar Received SIGINT")
        self.sigint_handler()

    def load_plugin(self, name, local=False):
        if local is True:
            plugin = __import__(name, fromlist=["."])
        else:
            plugin = __import__(
                "avatar2.plugins.%s" % name, fromlist=["avatar2.plugins"]
            )

        plugin.load_plugin(self)
        self.loaded_plugins += [name]

    @watch("AddTarget")
    def add_target(self, backend, *args, **kwargs):
        """
        Adds a new target to the analyses

        :ivar backend: the desired backend. Implemented for now: qemu, gdb
        :kwargs:       those argument will be passed to the target-object itself
        :return:       The created TargetObject
        """
        target = backend(self, *args, **kwargs)
        self.targets[target.name] = target

        return target

    def get_target(self, name):
        """
        Retrieves a target of the analyses by it's name

        :param name: The name of the desired target
        :return:     The Target!
        """

        return self.targets.get(name, None)

    def get_targets(self):
        """
        A generator for all targets.
        """
        for target in self.targets.items():
            yield target

    def init_targets(self):
        """
        Inits all created targets
        """
        for t in self.get_targets():
            t[1].init()

    def add_memory_range(
        self,
        address,
        size,
        name=None,
        permissions="rwx",
        file=None,
        file_offset=None,
        file_bytes=None,
        forwarded=False,
        forwarded_to=None,
        emulate=None,
        interval_tree=None,
        inline=False,
        overwrite=False,
        **kwargs
    ):
        """
        Adds a memory range to avatar

        :param emulate:      Emulation function that will take name, address and size if set
        :param inline:       If set to true, inline the emulation object into targets which support inlining (e.g., QEMU, PANDA).
                             NB: This deactivates all callbacks/watchmen on this region.
        :param address:      Base-Address of the Range
        :param size:         Size of the range
        :param file:         A file backing this range, if applicable
        :param file_offset:  The offset within the file
        :param file_bytes:   Bytes of the file to be copied into memory
        :param forwarded:    Whether this range should be forwarded
        :param forwarded_to: If forwarded is true, specify the forwarding target
        :param interval_tree:interval_tree this range shall be added to. If None,
                             the range will be added to self.memory_ranges
        :param overwrite:    If true, overwrite existing MR if they overlap with the new one
        """
        memory_ranges = self.memory_ranges if interval_tree is None else interval_tree
        if inline is True and emulate is None:
            self.log.warn("inline set for non peripheral memory")
        if emulate is not None:
            python_peripheral = emulate(name, address, size, **kwargs)
            forwarded = True
            forwarded_to = python_peripheral
            kwargs.update({"python_peripheral": python_peripheral})

        if forwarded is True:
            kwargs.update({"qemu_name": "avatar-rmemory"})
        if inline is True:
            kwargs.update({"qemu_name": "avatar-pyperipheral"})
            kwargs.update({"inline_module": str(python_peripheral.__module__)})

        m = MemoryRange(
            address,
            size,
            name=name,
            permissions=permissions,
            file=file,
            file_offset=file_offset,
            file_bytes=file_bytes,
            forwarded=forwarded,
            forwarded_to=forwarded_to,
            **kwargs
        )

        mr_set = memory_ranges[address : address + size]
        if overwrite is True and len(mr_set) > 0:
            start = min(mr_set, key=lambda x: x.begin).begin
            end = max(mr_set, key=lambda x: x.end).end
            memory_ranges.chop(
                address,
                address + size,
                datafunc=lambda x, y: deepcopy(x.data) if y is True else x.data,
            )
            mr_set2 = memory_ranges[start:end]
            for interval in mr_set2:
                interval.data.address = interval.begin
                interval.data.size = interval.end - interval.begin
                interval.data.name = "SPLIT_%x_%x" % (
                    interval.data.address,
                    interval.data.size,
                )

        memory_ranges[address : address + size] = m

        return m

    def get_memory_range(self, address):
        """
        Get a memory range from an address
        Note: for now just get's one range. If there are multiple ranges
        at the same address, this method won't work (for now)

        :param address: the address of the range
        :returns:       the memory range
        """
        ranges = self.memory_ranges[address]
        if len(ranges) > 1:
            raise Exception(
                "More than one memory range specified at 0x%x, \
                         not supported yet!"
                % address
            )
        elif len(ranges) == 0:
            self.log.critical("No Memory range specified at 0x%x" % address)
            return None
        return ranges.pop().data

    @watch("StateTransfer")
    def transfer_state(self, from_target, to_target, sync_regs=True, synced_ranges=[]):
        """
        Transfers the state from one target to another one

        :param from_target:     the source target
        :param to_target:       the destination target
        :param sync_regs:      Whether registers should be synced
        :param synced_ranges:  The memory ranges whose contents should be
                                transfered
        :type from_target:      Target()
        :type to_target:        Target()
        :type sync_regs:       bool
        :type synced_ranges:   list
        """

        if (
            from_target.state & TargetStates.STOPPED == 0
            or to_target.state & TargetStates.STOPPED == 0
        ):
            raise Exception(
                "Targets must be stopped for State Transfer, \
                             but target_states are (%s, %s)"
                % (from_target.state, to_target.state)
            )

        if sync_regs:
            # Test if we can take registers from TargetRegs-objects
            regs = (
                to_target.regs._get_names() & from_target.regs._get_names()
                if hasattr(to_target, "regs") and hasattr(from_target, "regs")
                else self.arch.registers
            )

            # ARM may have banked registers; Apparantly, the order in which we
            # write them is important to QEMU and could to lead bugs otherwise.
            if self.arch == ARM:
                regs = sorted(regs, key=lambda x: x[::-1])

            # The status register can cause a mode-switch, let's update it first
            if self.arch.sr_name in regs:
                regs = [self.arch.sr_name] + [r for r in regs if r != self.arch.sr_name]

            # Sync the registers!
            for r in regs:
                val = from_target.read_register(r)
                to_target.write_register(r, val)
            self.log.info("Synchronized Registers")

        for range in synced_ranges:
            m = from_target.read_memory(range.address, 1, range.size, raw=True)
            to_target.write_memory(range.address, 1, m, raw=True)
            self.log.info("Synchronized Memory Range: %s" % range.name)

    @watch("UpdateState")
    def _handle_update_state_message(self, message):
        self.log.info(
            "Received state update of target %s to %s"
            % (message.origin.name, message.state)
        )

    @watch("BreakpointHit")
    def _handle_breakpoint_hit_message(self, message):
        self.log.info("Breakpoint hit for Target: %s" % message.origin.name)
        self._handle_update_state_message(message)
        # Breakpoints are two stages: SYNCING | STOPPED -> HandleBreakpoint -> STOPPED
        # This makes sure that all handlers are complete before stopping and breaking wait()

        def bp_end_sync_cb(avatar, message, *args, **kwargs):
            avatar.watchmen.remove_watchman("BreakpointHit", w)
            avatar.fast_queue.put(
                UpdateStateMessage(message.origin, TargetStates.STOPPED)
            )

        # We handle this via a watchmen added in here, so we are sure that this
        # watchmen gets executed *at the end*
        # Note: This can break if another breakpoint-hit callback inserts an
        #       additional breakpointhit-watchmen (after).
        w = self.watchmen.add("BreakpointHit", when="after", callback=bp_end_sync_cb)

    @watch("SyscallCatched")
    def _handle_syscall_catched_message(self, message):
        self.log.info("Syscall catched for Target: %s" % message.origin.name)
        self._handle_update_state_message(message)

    @watch("RemoteMemoryRead")
    def _handle_remote_memory_read_message(self, message):

        range = self.get_memory_range(message.address)
        if not range:
            return (message.id, None, False)
        message.dst = range
        if not range.forwarded:
            raise Exception(
                "Forward request for non forwarded range received!\
                            (Address = 0x%x)"
                % message.address
            )
        if range.forwarded_to is None:
            raise Exception(
                "Forward request for non existing target received.\
                            (Address = 0x%x)"
                % message.address
            )

        try:
            kwargs = {"num_words": message.num_words, "raw": message.raw}
            if (
                hasattr(range.forwarded_to, "read_supports_pc")
                and range.forwarded_to.read_supports_pc is True
            ):
                kwargs["pc"] = message.pc

            mem = range.forwarded_to.read_memory(
                message.address, message.size, **kwargs
            )
            if not message.raw and message.num_words == 1 and not isinstance(mem, int):
                raise Exception(
                    (
                        "Forwarded read returned data of type %s "
                        "(expected: int)" % type(mem)
                    )
                )
            success = True
        except Exception as e:
            self.log.exception("RemoteMemoryRead failed: %s" % e)
            mem = -1
            success = False

        message.origin.protocols.remote_memory.send_response(message.id, mem, success)
        return (message.id, mem, success)

    @watch("RemoteMemoryWrite")
    def _handle_remote_memory_write_message(self, message):
        mem_range = self.get_memory_range(message.address)
        if not mem_range:
            message.origin.protocols.remote_memory.send_response(message.id, 0, True)
            return (message.id, 0, False)
        message.dst = mem_range
        if not mem_range.forwarded:
            raise Exception("Forward request for non forwarded range received!")
        if mem_range.forwarded_to is None:
            raise Exception("Forward request for non existing target received!")

        kwargs = {}
        if (
            hasattr(mem_range.forwarded_to, "write_supports_pc")
            and mem_range.forwarded_to.write_supports_pc is True
        ):
            kwargs["pc"] = message.pc

        success = mem_range.forwarded_to.write_memory(
            message.address, message.size, message.value, **kwargs
        )

        message.origin.protocols.remote_memory.send_response(message.id, 0, success)
        return message.id, 0, success

    def run(self):
        """
        The code of the Thread managing the asynchronous messages received.
        Default behavior: wait for the priority queue to hold a message and call
        the _async_handler method to process it.
        """
        self._close.clear()
        while True:
            if self._close.is_set():
                break

            try:
                message = self.queue.get(timeout=0.1)
            except:
                continue
            self.log.debug(
                "Avatar received %s. Queue-Status: %d/%d"
                % (message, self.queue.qsize(), self.fast_queue.qsize())
            )

            handler = self.message_handlers.get(message.__class__, None)
            if handler is None:
                raise Exception("No handler for Avatar-message %s registered" % message)
            else:
                handler(message)

    def stop(self):
        """
        Stop the thread which manages the asynchronous messages.
        """
        self._close.set()
        self.join()

    @watch("AvatarGetStatus")
    def get_status(self):
        return self.status


class AvatarFastQueueProcessor(Thread):
    """
    The avatar fast queue handles events which require immediate action,
    i.e. TargetStateUpdates.
    After processing, they get passed to the main avatar queue for further
    handling.
    """

    def __init__(self, avatar):
        super(AvatarFastQueueProcessor, self).__init__()
        self.avatar = avatar
        self._close = Event()
        self.daemon = True
        self.message_handlers = {
            UpdateStateMessage: self._fast_handle_update_state_message,
            BreakpointHitMessage: self._fast_handle_update_state_message,
            SyscallCatchedMessage: self._fast_handle_update_state_message,
        }

        self.start()

    def _fast_handle_update_state_message(self, message):
        # print message
        message.origin.update_state(message.state)
        self.avatar.queue.put(message)

    def run(self):
        self._close.clear()
        while True:
            if self._close.is_set():
                break

            # get() blocks sometimes.  This is a non-blocking wait.
            # if self.avatar.fast_queue.empty():
            # time.sleep(.001)
            # continue

            try:
                message = self.avatar.fast_queue.get(timeout=0.1)
            except:
                continue

            handler = self.message_handlers.get(message.__class__, None)
            if handler is None:
                raise Exception("No handler for fast message %s registered" % message)

            else:
                handler(message)

    def stop(self):
        """
        Stop the thread which manages the asynchronous messages.
        """
        self._close.set()
        self.join()



from types import MethodType
import os

from avatar2.watchmen import AFTER, BEFORE, watch
from avatar2 import TargetStates, GDBTarget


def dump_core_target(target, *args, **kwargs):
    """
    Stub method for compatibility to be able to call dump_core on the
    target object.
    """
    return dump_core(target.avatar, target, *args, **kwargs)


def dump_core(avatar, target, core_filename=""):
    """
    Create a core dump for the given GDB target
    :param core_filename: Filename to output core dump file to, can be absolute path or filename (to output to avatar's output directory); passing None causes the core dump to use default GDB naming scheme and is placed in the output directory.
    :return:              Tuple of return value, console output
    """
    if not isinstance(target, GDBTarget):
        raise TypeError("Core dumps can only be generated from GDBTargets")

    if core_filename == "":
        # If no core_filename name specifed, put file in output directory with
        # default GDB naming scheme.
        pid = 9999
        tokens = target.protocols.execution.console_command('info inferior')[1].split("\n")[2].split()
        for i in range(len(tokens)):
            if tokens[i] == "process":
                pid = tokens[i+1]
        
        # Return return value (True of False), and
        # output string of the console command.
        ret, output = target.protocols.execution.console_command('generate-core-file %s/core.%s' % (avatar.output_directory, pid))
        return ret, output
    elif os.path.dirname(core_filename) == "":
        # If only file basename provided, put core dump file in output directory.
        # Return return value (True of False), and
        # output string of the console command.
        return target.protocols.execution.console_command('generate-core-file %s/%s' % (avatar.output_directory, core_filename))
    elif os.path.isdir(os.path.dirname(core_filename)) == True:
        # Return return value (True of False), and
        # output string of the console command.
        return target.protocols.execution.console_command('generate-core-file %s' % core_filename)
    else:
        return False, ""


def add_methods(target):
    target.dump_core = MethodType(dump_core_target, target)


def target_added_callback(avatar, *args, **kwargs):
    target = kwargs["watched_return"]
    add_methods(target)


def load_plugin(avatar):
    avatar.watchmen.add_watchman(
        "AddTarget", when="after", callback=target_added_callback
    )
    avatar.dump_core = MethodType(dump_core, avatar)
    for target in avatar.targets.values():
        add_methods(target)



from threading import Thread
from time import sleep
from os.path import abspath
from avatar2.targets import PandaTarget

from ..watchmen import watch
from .target import action_valid_decorator_factory, TargetStates


class PyPandaTarget(PandaTarget):
    '''
    The pypanda target is a PANDA target, but uses pypanda to run the framework.

    '''
    def __init__(self, *args, **kwargs):
        try:
            import pandare
        except ImportError:
            raise RuntimeError(("PyPanda could not be found! for installation, "
                    "please follow the steps at https://github.com/"
                    "panda-re/panda/blob/master/panda/pypanda/docs/USAGE.md"))
        kwargs['executable'] = abspath(pandare.__file__)

        super(PyPandaTarget, self).__init__(*args, **kwargs)

        self.cb_ctx  = 0
        self.pypanda = None
        self._thread = None

    def shutdown(self):
        if self._thread is not None and self._thread.is_alive():
            self.protocols.execution.remote_disconnect()
            self.pypanda.end_analysis()

            # Wait for shutdown
            while self._thread.is_alive():
                sleep(.01)
                
        super(PyPandaTarget, self).shutdown()

    @watch('TargetInit')
    def init(self, **kwargs):
        from pandare import Panda

        arch = self.avatar.arch.qemu_name
        args = self.assemble_cmd_line()[1:]



        self.avatar.save_config(file_name=self.qemu_config_file,
                                config=self.generate_qemu_config())


        self.pypanda = Panda(arch=arch, extra_args=args, **kwargs)


        # adjust panda's signal handler to avatar2-standard
        def SigHandler(SIG,a,b):
            if self.state == TargetStates.RUNNING:
                self.stop()
                self.wait()

            self.avatar.sigint_handler()



        self.pypanda.setup_internal_signal_handler(signal_handler=SigHandler)

        self._thread = Thread(target=self.pypanda.run, daemon=True)
        self._thread.start()

        self._connect_protocols()



    def register_callback(self, callback, function, name=None, enabled=True,
                          procname=None):
        pp = self.pypanda

        if hasattr(pp.callback, callback) is False:
            raise Exception("Callback %s not found!" % callback)
        cb = getattr(pp.callback, callback)

        if name == None:
            name = 'avatar_cb_%d' % self.cb_ctx
        self.cb_ctx += 1

        pp.register_callback(cb, cb(function), name, enabled=enabled,
                             procname=procname)

        return name

    def disable_callback(self, name):
        pp = self.pypanda
        pp.disable_callback(name)

    def enable_callback(self, name):
        pp = self.pypanda
        pp.enable_callback(name)

    def add_hook(self, address, function, enabled=True,
                 kernel=True, asid=None, cb_type="before_block_exec"):
        '''
        This function registers hook at specified address with pypanda
        :param address: Address to be hooked.
        :param function: Function to be executed at specified address.
                         If the cb_type is "before_block_exec" (the default),
                         the arguments passed to that functions are cdata
                         pointer to the following structs:
                         cpustate *, TranslationBlock *, hook *
        '''
        self.pypanda.hook(address, enabled=enabled, kernel=kernel, asid=asid,
                          cb_type=cb_type)(function)


    @watch('TargetReadMemory')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
    def read_memory(self, address, size, num_words=1, raw=False):
        if raw == False:
            return self.protocols.memory.read_memory(address, size, num_words)
        else:
            return self.pypanda.physical_memory_read(address,size*num_words)


    @watch('TargetWriteMemory')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
    def write_memory(self, address, size, value, num_words=1, raw=False):
        if raw == False:
            return self.protocols.memory.write_memory(address, size, value, num_words=num_words)
        else:
            return self.pypanda.physical_memory_write(address, value)



    def delete_callback(self, name):
        return self.pypanda.delete_callback(name)




import logging

from os import O_WRONLY, O_RDONLY
from threading import Thread, Event, Condition
from ctypes import Structure, c_uint32, c_uint64
from enum import Enum
from posix_ipc import MessageQueue, ExistentialError

from avatar2.message import RemoteInterruptEnterMessage
from avatar2.message import RemoteInterruptExitMessage
from avatar2.targets import QemuTarget


class RINOperation(Enum):
    ENTER = 0
    EXIT = 1


class V7MRemoteInterruptNotification(Structure):
    _fields_ = [
        ('id', c_uint64),
        ('num_irq', c_uint32),
        ('operation', c_uint32),
        ('type', c_uint32)
    ]


class V7MInterruptNotificationAck(Structure):
    _fields_ = [
        ('id', c_uint64),
        ('success', c_uint32),
        ('operation', c_uint32),
    ]


class ARMV7MInterruptProtocol(Thread):
    """
    This protocol has two purposes: 
        a) injecting interrupts into an analysis target
        b) extracting interrupt exits and putting them into the avatar queue
    (b) is necessary in cases where two targets need to be synched on the
        interrupts.
        The way a v7m-nvic implements interrupt return is to put a magic value
        into $pc, and the hardware does the actual magic of popping from the
        interrupt stack and restoring the context. 
        However, the magic value defines the type of the interrupt return,
        and is hence synchronized on interrupt exit, alongside with the
        interrupt number
    :param origin:        Reference to the Target utilizing this protocol
    :param rx_queue_name: Name of the queue for receiving
    :param tx_queue_name: Name of the queue for sending
    """

    def __init__(self, origin, rx_queue_name, tx_queue_name):
        super(self.__class__, self).__init__()
        self._rx_queue_name = rx_queue_name
        self._tx_queue_name = tx_queue_name
        self._rx_queue = None
        self._tx_queue = None
        self._avatar_queue = origin.avatar.queue
        self._origin = origin
        self._close = Event()
        self._closed = Event()
        self._close.clear()
        self._closed.clear()
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def run(self):
        while True:
            if self._close.is_set():
                break

            request = None
            try:
                request = self._rx_queue.receive(0.5)
            except:
                continue

            req_struct = V7MRemoteInterruptNotification.from_buffer_copy(
                request[0])

            if RINOperation(req_struct.operation) == RINOperation.ENTER:
                msg = RemoteInterruptEnterMessage(self._origin, req_struct.id,
                                                  req_struct.num_irq)
            elif RINOperation(req_struct.operation) == RINOperation.EXIT:
                msg = RemoteInterruptExitMessage(self._origin, req_struct.id,
                                                 req_struct.type,
                                                 req_struct.num_irq)
                self.log.debug(
                    "Received an InterruptExitRequest for irq %d (%x)" %
                    (req_struct.num_irq, req_struct.type))

            else:
                msg = None
                raise Exception(("Received V7MRemoteInterrupt Notification with"
                                 "unknown operation type %d") %
                                req_struct.operation)

            self._avatar_queue.put(msg)

        self._closed.set()

    def stop(self):
        self._close.set()
        self._closed.wait()

    def enable_interrupts(self):
        if isinstance(self._origin, QemuTarget):
            # TODO: Make this more clean, i.e., check for remote memory
            rmem_rx_qname = self._origin.protocols.remote_memory.rx_queue_name
            rmem_tx_qname = self._origin.protocols.remote_memory.tx_queue_name
            # the tx-queue for qemu is the rx-queue for avatar and vice versa
            self._origin.protocols.monitor.execute_command(
                'avatar-armv7m-enable-irq',
                {'irq_rx_queue_name': self._tx_queue_name,
                 'irq_tx_queue_name': self._rx_queue_name,
                 'rmem_rx_queue_name': rmem_tx_qname,
                 'rmem_tx_queue_name': rmem_rx_qname
                 }
            )
        else:
            raise Exception("V7MInterruptProtocol is not implemented for %s" %
                            self._origin.__class__)

        try:
            self._rx_queue = MessageQueue(self._rx_queue_name, flags=O_RDONLY,
                                          read=True, write=False)
        except Exception as e:
            self.log.error("Unable to create rx_queue: %s" % e)
            return False

        try:
            self._tx_queue = MessageQueue(self._tx_queue_name, flags=O_WRONLY,
                                          read=False, write=True)
        except Exception as e:
            self.log.error("Unable to create tx_queue: %s" % e)
            self._rx_queue.close()
            return False

        self.daemon = True
        self.start()
        self.log.info("Enabled Interrupt Forwarding for %s" % self._origin)
        return True

    def ignore_interrupt_return(self, interrupt_number):
        if isinstance(self._origin, QemuTarget):
            self.log.info(
                "Disable handling of irq return for %d" % interrupt_number)
            self._origin.protocols.monitor.execute_command(
                'avatar-armv7m-ignore-irq-return',
                {'num_irq': interrupt_number}
            )

    def unignore_interrupt_return(self, interrupt_number):
        if isinstance(self._origin, QemuTarget):
            self.log.info(
                "Re-enable handling of irq return for %d" % interrupt_number)
            self._origin.protocols.monitor.execute_command(
                'avatar-armv7m-unignore-irq-return',
                {'num_irq': interrupt_number}
            )

    def inject_interrupt(self, interrupt_number, cpu_number=0):
        if isinstance(self._origin, QemuTarget):
            self.log.info("Injecting interrupt %d" % interrupt_number)
            self._origin.protocols.monitor.execute_command(
                'avatar-armv7m-inject-irq',
                {'num_irq': interrupt_number, 'num_cpu': cpu_number}
            )

    def set_vector_table_base(self, base, cpu_number=0):
        if isinstance(self._origin, QemuTarget):
            self.log.info("Setting vector table base to 0x%x" % base)
            self._origin.protocols.monitor.execute_command(
                'avatar-armv7m-set-vector-table-base',
                {'base': base, 'num_cpu': cpu_number}
            )

    def send_interrupt_exit_response(self, id, success):
        response = V7MInterruptNotificationAck(id, success,
                                               RINOperation.EXIT.value)
        
        try:
            self._tx_queue.send(response)
            self.log.debug("Send RemoteInterruptExitResponse with id %d" % id)
            return True
        except Exception as e:
            self.log.error("Unable to send response: %s" % e)
            return False

    def send_interrupt_enter_response(self, id, success):
        response = V7MInterruptNotificationAck(id, success,
                                               RINOperation.ENTER.value)
        try:
            self._tx_queue.send(response)
            self.log.debug("Send RemoteInterruptEnterResponse with id %d" % id)
            return True
        except Exception as e:
            self.log.error("Unable to send response: %s" % e)
            return False

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        self.stop()
        if self._rx_queue:
            try:
                self._rx_queue.unlink()
                self._rx_queue.close()
                self._rx_queue = None
            except ExistentialError:
                self.log.warning("Tried to close/unlink non existent rx_queue")
        if self._tx_queue:
            try:
                self._tx_queue.unlink()
                self._tx_queue.close()
                self._tx_queue = None
            except ExistentialError:
                self.log.warning("Tried to close/unlink non existent tx_queue")



from capstone import *
from keystone.keystone_const import *
from unicorn import *
from unicorn.mips_const import *
from .architecture import Architecture
import avatar2

from avatar2.installer.config import QEMU, PANDA, OPENOCD, GDB_MULTI


class MIPS32(Architecture):

    get_qemu_executable = Architecture.resolve(QEMU)
    get_panda_executable = Architecture.resolve(PANDA)
    get_gdb_executable  = Architecture.resolve(GDB_MULTI)
    get_oocd_executable = Architecture.resolve(OPENOCD)


    registers = {
        'zero': 0, 
        'at': 1, 
        'v0': 2, 'v1': 3, 
        'a0': 4, 'a1': 5, 'a2': 6, 'a3': 7, 
        't0': 8, 't1': 9, 't2': 10, 't3': 11, 't4': 12, 't5': 13, 't6': 14, 't7': 15, 
        's0': 16, 's1': 17, 's2': 18, 's3': 19, 's4': 20, 's5': 21, 's6': 22, 's7': 23, 
        't8': 24, 't9': 25, 
        'k0': 26, 'k1': 27,
        'gp': 28, 'sp': 29, 'fp': 30, 'ra': 31, 'pc': 32,
    }

    unicorn_registers = {
        'zero': UC_MIPS_REG_0, 'at': UC_MIPS_REG_1, 'v0': UC_MIPS_REG_2,
        'v1': UC_MIPS_REG_3, 'a0': UC_MIPS_REG_4, 'a1': UC_MIPS_REG_5,
        'a2': UC_MIPS_REG_6, 'a3': UC_MIPS_REG_7, 't0': UC_MIPS_REG_8,
        't1': UC_MIPS_REG_9, 't2': UC_MIPS_REG_10, 't3': UC_MIPS_REG_11,
        't4': UC_MIPS_REG_12, 't5': UC_MIPS_REG_13, 't6': UC_MIPS_REG_14,
        't7': UC_MIPS_REG_15, 's0': UC_MIPS_REG_16, 's1': UC_MIPS_REG_17,
        's2': UC_MIPS_REG_18, 's3': UC_MIPS_REG_19, 's4': UC_MIPS_REG_20,
        's5': UC_MIPS_REG_21, 's6': UC_MIPS_REG_22, 's7': UC_MIPS_REG_23,
        't8': UC_MIPS_REG_24, 't9': UC_MIPS_REG_25, 'k0': UC_MIPS_REG_26,
        'k1': UC_MIPS_REG_27, 'gp': UC_MIPS_REG_28, 'sp': UC_MIPS_REG_29,
        'fp': UC_MIPS_REG_30, 'ra': UC_MIPS_REG_31, 'pc': UC_MIPS_REG_PC
        }

    pc_name = 'pc'
    sr_name = 'cpsr'

    capstone_arch = CS_ARCH_MIPS
    keystone_arch = KS_ARCH_MIPS

    unicorn_arch = UC_ARCH_MIPS
    unicorn_mode = UC_MODE_MIPS32


class MIPS_BE(MIPS32):

    endian = 'big'

    qemu_name = 'mips'
    gdb_name = 'mips'
    angr_name = 'mips'

    capstone_mode = CS_MODE_BIG_ENDIAN + CS_MODE_MIPS32
    keystone_mode = KS_MODE_BIG_ENDIAN + KS_MODE_MIPS32
    unicorn_mode = UC_MODE_BIG_ENDIAN | UC_MODE_THUMB


class MIPS_LE(MIPS32):

    endian = 'little'

    qemu_name = 'mipsel'
    gdb_name = 'mips'
    angr_name = 'mipsel'

    capstone_mode = CS_MODE_LITTLE_ENDIAN + CS_MODE_MIPS32
    keystone_mode = KS_MODE_LITTLE_ENDIAN + KS_MODE_MIPS32
    unicorn_mode = UC_MODE_LITTLE_ENDIAN | UC_MODE_THUMB



class MIPS_24KF(MIPS_BE):

    cpu_model = '24Kf'




from avatar2.targets import QemuTarget
from avatar2.targets import TargetStates
from avatar2.targets import action_valid_decorator_factory


class PandaTarget(QemuTarget):
    def __init__(self, *args, **kwargs):
        super(PandaTarget, self).__init__(*args, **kwargs)

        executable = kwargs.get('executable')
        self.executable = (executable if executable is not None
                           else self._arch.get_panda_executable())

        # self.protocols.monitor = self.protocols.execution

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def begin_record(self, record_name):
        """
        Starts recording the execution in PANDA

        :param record_name:   The name of the record file
        """
        filename = "%s/%s" % (self.avatar.output_directory, record_name)
        return self.protocols.monitor.execute_command('begin_record',
                                                      {'file_name': filename})
        # self.protocols.monitor._sync_request('monitor begin_record "%s"'
        #                                    % filename, 'done')

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def end_record(self):
        """
        Stops recording the execution in PANDA
        """
        return self.protocols.monitor.execute_command('end_record')
        # self.protocols.monitor._sync_request('monitor end_record', 'done')

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def begin_replay(self, replay_name, cont=True):
        """
        Starts replaying a captured replay

        :param replay_name: The name of the file to be replayed
        :param cont: Whether execution shall automatically be resumed (default True)
        """
        self.protocols.monitor.execute_command('begin_replay',
                                               {'file_name': replay_name})
        if cont is True:
            self.cont()

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def end_replay(self):
        """
        Stops a current ongoing replay
        """
        return self.protocols.monitor.execute_command('end_replay')

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def load_plugin(self, plugin_name, plugin_args=None, file_name=None):
        """
        Loads a PANDA plugin

        :param plugin_name: The name of the plugin to be loaded
        :param plugin_args: Arguments to be passed to the plugin,
                            aseperated by commas
        :param file_name:   Absolute path to the plugin shared object file,
                            in case that the default one should not be used
        """

        args_dict = {'plugin_name': plugin_name}
        if plugin_args:
            args_dict['plugin_args'] = plugin_args
        if file_name:
            args_dict['file_name'] = file_name

        return self.protocols.monitor.execute_command('load_plugin', args_dict)

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def unload_plugin(self, plugin_name):
        """
        Unloads a PANDA plugin

        :param plugin_name: The name of the plugin to be unloaded
        :return: True if the requested plugin was present
        """
        full_plugin_name = 'panda_%s.so' % plugin_name
        for plugin_dict in self.list_plugins():
            if plugin_dict['name'] == full_plugin_name:
                self.protocols.monitor.execute_command('unload_plugin',
                                                       {'index': plugin_dict['index']})
                return True
        return False

    @action_valid_decorator_factory(TargetStates.STOPPED, 'monitor')
    def list_plugins(self):
        """
        Lists the laoded PANDA plugins

        :return: a list with the loaded panda_plugins
        """
        return self.protocols.monitor.execute_command('list_plugins')



import sys

if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

import logging
import json
import telnetlib
import re

from avatar2 import archs


class QMPProtocol(object):
    def __init__(self, port, origin=None):

        self.port = port
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)
        self.origin=origin
        self.id = 0

    def __del__(self):
        self.shutdown()

    def connect(self):
        self._telnet = telnetlib.Telnet('127.0.0.1', self.port)
        self._telnet.read_until('\r\n'.encode('ascii'))
        self.execute_command('qmp_capabilities')
        return True

    def execute_command(self, cmd, args=None):
        command = {}
        command['execute'] = cmd
        if args:
            command['arguments'] = args
        command['id'] = self.id
        self._telnet.write(('%s\r\n' % json.dumps(command)).encode('ascii'))

        while True:
            resp = self._telnet.read_until('\r\n'.encode('ascii'))
            self.log.info(resp)
            resp = json.loads(resp.decode('ascii'))
            if 'event' in resp:
                continue
            if 'id' in resp:
                break
        if resp['id'] != self.id:
            raise Exception('Mismatching id for qmp response')
        self.id += 1
        if 'error' in resp:
            return resp['error']
        if 'return' in resp:
            return resp['return']
        raise Exception("Response contained neither an error nor an return")

    def reset(self):
        """
        Resets the target
        returns: True on success, else False
        """
        pass

    def shutdown(self):
        """
        returns: True on success, else False
        """
        # self._communicator.stop()
        pass

    def get_registers(self, raw=False):
        """
        Gets the current register state based on the hmp info registers
        command. In comparison to register-access with the register protocol,
        this function can also be called while the target is executing.
        :param raw:     If true, return output of the hmp-cmd without parsing
        :type raw:      ``bool``
        :return:        A dictionary with the registers or the register-string
        """

        regs_s = self.execute_command("human-monitor-command",
                                      {"command-line": "info registers"})
        if raw is True:
            return regs_s

        if issubclass(self.origin.avatar.arch, archs.ARM):
            regs_r = re.findall('(...)=([0-9a-f]{8})', regs_s)
            return dict([(r.lower(), int(v, 16)) for r, v in regs_r])
        else:
            raise Exception( ("get_registers in non-raw mode called on "
                              "unsupported arch %s") %
                              self.origin.avatar.arch.__name__)

    def x86_get_segment_register_base(self, reg):
        """
        x86-only: get the segment register base using the hmp.
        :param reg:     name of the register
        :type reg:      ``str``
        :return:        The base of segment register ``reg`` on success
        """

        regs = self.get_registers(raw=True)
        regs_r = re.findall(reg.upper()+'\s?=[0-9a-f]+\s([0-9a-f]+)', regs)
        if len(regs_r) != 1:
            raise exception("Couldn't retrieve base for %s" % reg)

        return int(regs_r[0], 16)



#from __future__ import absolute_import

#from . import archs

from .memory_range import MemoryRange
from .message import *
from .archs import *
from .protocols import *
from .targets import *
from .avatar2 import Avatar
#from .analyses import Analyses

#GDB_TARGET = targets.gdb.GDBTarget
#QEMU_TARGET = targets.qemu.QemuTarget



# from capstone import CS_ARCH_ARM, CS_MODE_LITTLE_ENDIAN, CS_MODE_BIG_ENDIAN

from capstone import *
from keystone.keystone_const import *
from unicorn import *
from unicorn.arm_const import *
from .architecture import Architecture
import avatar2

from avatar2.installer.config import QEMU, PANDA, OPENOCD, GDB_MULTI

class ARM(Architecture):

    get_qemu_executable = Architecture.resolve(QEMU)
    get_panda_executable = Architecture.resolve(PANDA)
    get_gdb_executable  = Architecture.resolve(GDB_MULTI)
    get_oocd_executable = Architecture.resolve(OPENOCD)



    qemu_name = 'arm'
    gdb_name = 'arm'
    registers = {'r0': 0, 'r1': 1, 'r2': 2, 'r3': 3, 'r4': 4, 'r5': 5, 'r6': 6,
                 'r7': 7, 'r8': 8, 'r9': 9, 'r10': 10, 'r11': 11, 'r12': 12,
                 'sp': 13, 'lr': 14, 'pc': 15, 'cpsr': 25,
                 }
    unicorn_registers = {'r0': UC_ARM_REG_R0, 'r1': UC_ARM_REG_R1, 'r2': UC_ARM_REG_R2,
                         'r3': UC_ARM_REG_R3, 'r4': UC_ARM_REG_R4, 'r5': UC_ARM_REG_R5,
                         'r6': UC_ARM_REG_R6, 'r7': UC_ARM_REG_R7, 'r8': UC_ARM_REG_R8,
                         'r9': UC_ARM_REG_R9, 'r10': UC_ARM_REG_R10, 'r11': UC_ARM_REG_R11,
                         'r12': UC_ARM_REG_R12, 'sp': UC_ARM_REG_SP, 'lr': UC_ARM_REG_LR,
                         'pc': UC_ARM_REG_PC, 'cpsr': UC_ARM_REG_CPSR}
    pc_name = 'pc'
    sr_name = 'cpsr'
    unemulated_instructions = ['mcr', 'mrc']
    capstone_arch = CS_ARCH_ARM
    capstone_mode = CS_MODE_LITTLE_ENDIAN
    keystone_arch = KS_ARCH_ARM
    keystone_mode = KS_MODE_ARM
    unicorn_arch = UC_ARCH_ARM
    unicorn_mode = UC_MODE_ARM

class ARM_CORTEX_M3(ARM):
    cpu_model = 'cortex-m3'
    qemu_name = 'arm'
    gdb_name = 'arm'

    capstone_arch = CS_ARCH_ARM
    keystone_arch = KS_ARCH_ARM
    capstone_mode = CS_MODE_LITTLE_ENDIAN | CS_MODE_THUMB | CS_MODE_MCLASS
    keystone_arch = KS_ARCH_ARM
    keystone_mode = KS_MODE_LITTLE_ENDIAN | KS_MODE_THUMB
    unicorn_arch = UC_ARCH_ARM
    unicorn_mode = UC_MODE_LITTLE_ENDIAN | UC_MODE_THUMB
    sr_name = 'xpsr'


    @staticmethod
    def register_write_cb(avatar, *args, **kwargs):
                
        if isinstance(kwargs['watched_target'],
                      avatar2.targets.qemu_target.QemuTarget):
            qemu = kwargs['watched_target']

            # xcps/cpsr encodes the thumbbit diffently accross different
            # ISA versions. Panda_target does not cleanly support cortex-m yet,
            # and hence uses the thumbbit as stored on other ARM versions.
            if isinstance(qemu, avatar2.targets.panda_target.PandaTarget):
                shiftval = 5
            else:
                shiftval = 24

            if args[0] == 'pc' or args[0] == 'cpsr':
                cpsr = qemu.read_register('cpsr')
                if cpsr & 1<< shiftval:
                    return
                else:
                    cpsr |= 1<<shiftval
                    qemu.write_register('cpsr', cpsr)

    @staticmethod
    def init(avatar):
        avatar.watchmen.add('TargetRegisterWrite', 'after',
                            ARM_CORTEX_M3.register_write_cb)

        pass
ARMV7M = ARM_CORTEX_M3


class ARMBE(ARM):
    qemu_name = 'armeb'
    capstone_mode = CS_MODE_BIG_ENDIAN



from types import MethodType

import logging

from avatar2.watchmen import AFTER, BEFORE, watch
from avatar2.archs import ARM, ARMBE

from cmsis_svd.parser import SVDParser

def load_svd(self, vendor, file):
    parser = SVDParser.for_packaged_svd(vendor, file)
    for peripheral in parser.get_device().peripherals:
        p_dict = peripheral.to_dict()
        base_addr = p_dict['base_address']
        size = int(p_dict['address_block']['size']) -1 #hackhack
        name = peripheral.name
        self.add_memory_range(base_addr, size)



def load_plugin(avatar):
    if avatar.arch not in [ARM, ARMBE]:
        log = logging.getLogger('avatar')
        log.warning("SVD-reader requires an ARM core")

    avatar.load_svd = MethodType(load_svd, avatar)



import sys

from intervaltree import IntervalTree
from inspect import signature

if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    from functools import cached_property
else:
    from cached_property import cached_property



class AvatarPeripheral(object):
    def __init__(self, name, address, size, **kwargs):
        self.name = name if name else "%s_%x" % (self.__class__.__name__, address)
        self.address = address
        self.size = size
        self.read_handler = IntervalTree()
        self.write_handler = IntervalTree()

    @cached_property
    def read_supports_pc(self):
        """
        Checks if all registered read-handlers support an pc parameter
        """
        return all("pc" in signature(rh.data).parameters for rh in self.read_handler)

    @cached_property
    def write_supports_pc(self):
        """
        Checks if all registered read-handlers support an pc parameter
        """
        return all("pc" in signature(rh.data).parameters for rh in self.write_handler)

    def shutdown(self):
        """
        Some peripherals will require to be shutdowned when avatar exits.
        In those cases, this method should be overwritten.
        """
        pass

    def write_memory(self, address, size, value, num_words=1, raw=False, pc=0):

        if num_words != 1 or raw is True:
            raise Exception(
                "write_memory for AvatarPeripheral does not support \
                             'num_words' or 'raw' kwarg"
            )

        offset = address - self.address
        intervals = self.write_handler[offset : offset + size]
        if intervals == set():
            raise Exception(
                "No write handler for peripheral %s at offset %d \
                            (0x%x)"
                % (self.name, offset, address)
            )
        if len(intervals) > 1:
            raise Exception(
                "Multiple write handler for peripheral %s\
                            at offset %d"
                % (self.name, offset)
            )

        kwargs = {} if self.write_supports_pc is False else {"pc": pc}
        return intervals.pop().data(offset, size, value, **kwargs)

    def read_memory(self, address, size, num_words=1, raw=False, pc=0):
        if num_words != 1 or raw is True:
            raise Exception(
                "read_memory for AvatarPeripheral does not support \
                             'num_words' or 'raw' kwarg"
            )

        offset = address - self.address
        intervals = self.read_handler[offset : offset + size]

        if intervals == set():
            raise Exception(
                "No read handler for peripheral %s at offset %d \
                            (0x%x)"
                % (self.name, offset, address)
            )
        if len(intervals) > 1:
            raise Exception(
                "Multiple read handler for peripheral %s\
                            at offset %d"
                % (self.name, offset)
            )
        kwargs = {} if self.write_supports_pc is False else {"pc": pc}
        return intervals.pop().data(offset, size, **kwargs)



from avatar2.targets import Target, TargetStates
from avatar2.protocols.gdb import GDBProtocol

from .target import action_valid_decorator_factory, synchronize_state
from ..watchmen import watch

class GDBTarget(Target):
    def __init__(self, avatar,
                 gdb_executable=None, gdb_additional_args=None, 
                 gdb_ip='127.0.0.1', gdb_port=3333,
                 gdb_serial_device='/dev/ttyACM0',
                 gdb_serial_baud_rate=38400,
                 gdb_serial_parity='none',
                 gdb_verbose_mi=False,
                 serial=False,
                 enable_init_files=False,
                 local_binary=None,
                 arguments=None,
                 **kwargs
                 ):

        super(GDBTarget, self).__init__(avatar, **kwargs)

        self.gdb_executable = (gdb_executable if gdb_executable is not None
                               else self._arch.get_gdb_executable())
        self.gdb_additional_args = gdb_additional_args if gdb_additional_args else []
        self.gdb_ip = gdb_ip
        self.gdb_port = gdb_port
        self.gdb_serial_device = gdb_serial_device
        self.gdb_serial_baud_rate = gdb_serial_baud_rate
        self.gdb_serial_parity = gdb_serial_parity
        self._serial = serial
        self._local_binary = local_binary
        self._arguments = arguments
        self._enable_init_files = enable_init_files
        self._verbose_gdbmi = gdb_verbose_mi

    def init(self):

        gdb = GDBProtocol(gdb_executable=self.gdb_executable,
                          arch=self._arch,
                          additional_args=self.gdb_additional_args,
                          avatar=self.avatar, origin=self,
                          enable_init_files=self._enable_init_files,
                          binary=self._local_binary,
                          local_arguments=self._arguments,
                          verbose=self._verbose_gdbmi)

        # If we are debugging a program locally,
        # we do not need to establish any connections
        if not self._local_binary:
            if not self._serial:
                if gdb.remote_connect(ip=self.gdb_ip, port=self.gdb_port):
                    self.log.info("Connected to Target")
                else:
                    self.log.warning("Connecting failed")
            else:
                if gdb.remote_connect_serial(device=self.gdb_serial_device,
                                             baud_rate=self.gdb_serial_baud_rate,
                                             parity=self.gdb_serial_parity):
                    self.log.info("Connected to Target")
                else:
                    self.log.warning("Connecting failed")
        else:
            self.update_state(TargetStates.INITIALIZED)

        self.protocols.set_all(gdb)

        if self._local_binary:
            self.wait(state=TargetStates.INITIALIZED)
        else:
            self.wait()

    @watch('TargetCont')
    @action_valid_decorator_factory(TargetStates.INITIALIZED, 'execution')
    @synchronize_state(TargetStates.RUNNING)
    def run(self, blocking=True):
        return self.protocols.execution.run()

    def cont(self, blocking=True):
        if self.state != TargetStates.INITIALIZED:
            super(GDBTarget, self).cont(blocking=blocking)
        else:
            self.run()

    @action_valid_decorator_factory(TargetStates.INITIALIZED, 'execution')
    def disable_aslr(self):
        self.protocols.execution.set_gdb_variable('disable-randomization',
                                                  'on')

    @action_valid_decorator_factory(TargetStates.INITIALIZED, 'execution')
    def enable_aslr(self):
        self.protocols.execution.set_gdb_variable('disable-randomization',
                                                  'off')
    



import socket

from threading import Thread, Lock, Event
from .avatar_peripheral import AvatarPeripheral

SR_RXNE = 0x20
SR_TXE = 0x80
SR_TC = 0x40


class NucleoRTC(AvatarPeripheral):
    def nop_read(self, offset, size):
        return 0x00

    def __init__(self, name, address, size, **kwargs):
        AvatarPeripheral.__init__(self, name, address, size)
        self.read_handler[0:size] = self.nop_read


class NucleoTIM(AvatarPeripheral):
    def nop_read(self, offset, size):
        return 0x00

    def nop_write(self, offset, size, value):
        return True

    def __init__(self, name, address, size, **kwargs):
        AvatarPeripheral.__init__(self, name, address, size)
        self.read_handler[0:size] = self.nop_read
        self.write_handler[0:size] = self.nop_write


class NucleoUSART(AvatarPeripheral, Thread):
    def read_status_register(self, offset, size):
        self.lock.acquire(True)
        ret = self.status_register
        self.lock.release()
        return ret

    def read_data_register(self, offset, size):
        self.lock.acquire(True)
        ret = self.data_buf[0]
        self.data_buf = self.data_buf[1:]
        if len(self.data_buf) == 0:
            self.status_register &= ~SR_RXNE
        self.lock.release()
        #print(">>> %s" % hex(ret))
        return ret

    def write_data_register(self, offset, size, value):
        if self.connected:
            self.conn.send(bytes((chr(value).encode('utf-8'))))
        #print("<<< %s" % hex(value))
        return True

    def nop_read(self, offset, size):
        return 0x00

    def nop_write(self, offset, size, value):
        return True

    def __init__(self, name, address, size, nucleo_usart_port=5656, **kwargs):
        Thread.__init__(self)
        AvatarPeripheral.__init__(self, name, address, size)
        self.port = nucleo_usart_port

        self.data_buf = bytearray()
        self.status_register = SR_TXE | SR_TC

        self.read_handler[0:4] = self.read_status_register
        self.read_handler[4:8] = self.read_data_register
        self.write_handler[0:4] = self.nop_write
        self.write_handler[4:8] = self.write_data_register

        self.read_handler[8:size] = self.nop_read
        self.write_handler[8:size] = self.nop_write

        self.connected = False

        self.lock = Lock()
        self._close = Event()
        self.sock = None
        self.conn = None
        self.daemon = True
        self.start()

    def shutdown(self):
        self._close.set()

        if self.conn:
            self.conn.close()

        if self.sock:
            self.sock.close()

    # Pretender compat layer.
    def read(self, address, size):
        return self.read_memory(address, size)

    def write(self, address, size, value):
        return self.write_memory(address, size, value)

    def run(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.setblocking(0)
        self.sock.bind(('127.0.0.1', self.port))
        self.sock.settimeout(0.1)

        while not self._close.is_set():
            self.sock.listen(1)

            try:
                self.conn, addr = self.sock.accept()
                self.conn.settimeout(0.1)
                self.connected = True
            except socket.timeout:
                continue
            except OSError as e:
                if e.errno == 9:
                    # Bad file descriptor error. Only happens when we called
                    # closed on the socket, the continuing the loop will 
                    # terminate, which is the desired behaviour
                    continue
                else:
                    # Something terrible happened
                    raise (e)

            while not self._close.is_set():
                try:
                    chr = self.conn.recv(1)
                except socket.timeout:
                    continue
                if not chr:
                    break
                self.lock.acquire(True)
                self.data_buf += chr
                self.status_register |= SR_RXNE
                self.lock.release()
            self.connected = False



from types import MethodType

from keystone import *


def assemble(self, asmstr, addr=None,
                arch=None, mode=None):
    """
    Main purpose of the assembler plugin, it's used to assemble
    instructions
    :param addr:  The address to start disassembling.
                  If not specified, the current pc is used
    :param arch:  The keystone-architecture to be used.
    :param mode:  The keystone-mode to be used.
                  If not specified, it is retrieved from avatar.arch
    :returns:     Raw bytes
    """

    arch = self._arch.keystone_arch if not arch else arch
    mode = self._arch.keystone_mode if not mode else mode
    addr = self.regs.pc if not addr else addr

    md = Ks(arch, mode)
    bytelist = md.asm(asmstr, addr)[0]
    bytes_raw = bytes(bytelist)
    return bytes_raw

def inject_asm(self, asmstr, addr=None, arch=None, mode=None):
    """
    Assemble the string, and inject it into the target)
    """
    arch = self._arch.keystone_arch if not arch else arch
    mode = self._arch.keystone_mode if not mode else mode
    addr = self.regs.pc if not addr else addr

    md = Ks(arch, mode)
    bytelist = md.asm(asmstr, addr)[0]
    bytes_raw = bytes(bytelist)
    return self.write_memory(addr, 1, bytes_raw, len(bytes_raw), raw=True)

def target_added_callback(avatar, *args, **kwargs):
    target = kwargs['watched_return']
    target.assemble = MethodType(assemble, target)
    target.inject_asm = MethodType(inject_asm, target)


def load_plugin(avatar):
    avatar.watchmen.add_watchman('AddTarget', when='after',
                                 callback=target_added_callback)
    for target in avatar.targets.values():
        target.assemble = MethodType(assemble, target)
        target.inject_asm = MethodType(inject_asm, target)



import sys
from avatar2.targets import Target, TargetStates
from avatar2.protocols.jlink import JLinkProtocol
from avatar2.watchmen import watch

if sys.version_info < (3, 0):
    from Queue import PriorityQueue
else:
    from queue import PriorityQueue


class JLinkTarget(Target):
    def __init__(self, avatar, serial, device, **kwargs):
        """
        Create a JLink target instance
        :param avatar: The avatar instance
        :param serial: The JLink's serial number
        :param device: The Device string to use (e.g., ARM7, see JlinkExe for the list)
        :param kwargs:
        """
        super(JLinkTarget, self).__init__(avatar, **kwargs)
        self.avatar = avatar
        self.serial = serial
        self.device = device

    @watch("TargetInit")
    def init(self):
        jlink = JLinkProtocol(serial=self.serial, device=self.device, avatar=self.avatar, origin=self)
        self.protocols.set_all(jlink)
        if jlink.jlink.halted():
            self.state = TargetStates.STOPPED
        else:
            self.state = TargetStates.RUNNING
        #self.wait()

    def reset(self, halt=True):
        self.protocols.execution.reset(halt=halt)


from avatar2.installer.config import AvatarConfig
import distutils
from os import environ
from re import sub

class Architecture(object):
    
    @staticmethod 
    def _resolve_executable(exec_name):
        """
        Resolves the name of the executable for the endpoint, using the config
        file generated by avatar2's installer
        Convienently enough, this config already checks whether the executable
        is installed on the system
        """

        env_var_name = 'AVATAR2_%s_EXECUTABLE' % sub(r'avatar-|\W\(.*\)',
                                                     '', exec_name).upper()
        env_exec = environ.get( env_var_name )
        if env_exec is not None:
            target_path = distutils.spawn.find_executable(env_exec)
        else:
            target_path = AvatarConfig().get_target_path(exec_name)
        if target_path is None:
            raise Exception(("Couldn't find executable for %s\n"
                             "Have you tried running the avatar2-installer "
                             "(python -m avatar2.installer) or setting the %s "
                             "environment variable?" % (exec_name, env_var_name)
                            ))
        return target_path

    @staticmethod
    def resolve(exec_name):
        """
        This wrapper around _resolve_executable allows us to have targets
        only resolved when needed, and not when the arch-class gets imported
        """
        return staticmethod(lambda : Architecture._resolve_executable(exec_name))



    @staticmethod
    def init(avatar):
        pass
    
    registers = None
    special_registers = {}
    sr_name = None
